from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
import math
import re
from typing import Any, Optional, Union

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    num_to_chars,
    try_float,
    try_float_or_nan,
    try_float_or_none,
    try_int,
    try_int_or_none,
    try_str_from_series,
    try_str_from_series_or_none,
)

BLOCKS_LINE_REGEX = r"^##BLOCKS=\s*(\d+)$"
END_LINE_REGEX = "~End"
EXPORT_VERSION = "1.3"


def try_non_nan_float_or_none(value: Optional[str]) -> Optional[float]:
    number = try_float_or_none(value)
    return None if number is None or math.isnan(number) else number


def can_parse_as_float_non_nan(value: Any) -> bool:
    try:
        number = float(value)
    except ValueError:
        return False
    return not math.isnan(number)


def rm_df_columns(data: pd.DataFrame, pattern: str) -> pd.DataFrame:
    return data.drop(
        columns=[column for column in data.columns if re.match(pattern, column)]
    )


def try_str_from_series_multikey_or_none(
    data: pd.Series[Any],
    possible_keys: set[str],
) -> Optional[str]:
    for key in possible_keys:
        value = try_str_from_series_or_none(data, key)
        if value is not None:
            return value
    return None


def try_str_from_series_multikey(
    data: pd.Series[Any],
    possible_keys: set[str],
    msg: Optional[str] = None,
) -> str:
    return assert_not_none(
        try_str_from_series_multikey_or_none(data, possible_keys),
        msg=msg,
    )


NUM_WELLS_TO_PLATE_DIMENSIONS: dict[int, tuple[int, int]] = {
    6: (3, 2),
    12: (4, 3),
    24: (6, 4),
    48: (8, 6),
    96: (12, 8),
    384: (24, 16),
    1536: (48, 32),
}


def num_wells_to_n_columns(well_count: int) -> int:
    if dimensions := NUM_WELLS_TO_PLATE_DIMENSIONS.get(well_count):
        return dimensions[0]

    num_wells = ",".join(
        [str(num_wells) for num_wells in NUM_WELLS_TO_PLATE_DIMENSIONS]
    )
    msg = (
        f"Unknown number of wells '{well_count}'. Only accepted values are {num_wells}"
    )
    raise AllotropeConversionError(msg)


class ReadType(Enum):
    ENDPOINT = "Endpoint"
    KINETIC = "Kinetic"
    SPECTRUM = "Spectrum"
    WELL_SCAN = "Well Scan"


class ExportFormat(Enum):
    TIME_FORMAT = "TimeFormat"
    PLATE_FORMAT = "PlateFormat"


class DataType(Enum):
    RAW = "Raw"
    REDUCED = "Reduced"
    BOTH = "Both"


class ScanPosition(Enum):
    BOTTOM = "Bottom"
    TOP = "Top"
    NONE = None


@dataclass(frozen=True)
class Block:
    block_type: str


@dataclass(frozen=True)
class GroupDataElementEntry:
    name: str
    value: float


@dataclass(frozen=True)
class GroupDataElement:
    sample: str
    position: str
    plate: str
    entries: list[GroupDataElementEntry]


@dataclass(frozen=True)
class GroupSampleData:
    identifier: str
    data_elements: list[GroupDataElement]
    aggregated_entries: list[GroupDataElementEntry]

    @staticmethod
    def create(data: pd.DataFrame) -> GroupSampleData:
        top_row = data.iloc[0]
        identifier = str(top_row["Sample"])
        data = rm_df_columns(data, r"^Sample$|^Standard Value|^R$|^Unnamed: \d+$")
        numeric_columns = [
            column
            for column in data.columns
            if can_parse_as_float_non_nan(top_row[column])
        ]

        normal_columns = []
        aggregated_columns = []
        for column in numeric_columns:
            if data[column].iloc[1:].isnull().all():
                aggregated_columns.append(column)
            else:
                normal_columns.append(column)

        return GroupSampleData(
            identifier=identifier,
            data_elements=[
                GroupDataElement(
                    sample=identifier,
                    position=try_str_from_series_multikey(
                        row,
                        {"Well", "Wells"},
                        msg="Unable to find well position in group data.",
                    ),
                    plate=try_str_from_series(row, "WellPlateName"),
                    entries=[
                        GroupDataElementEntry(
                            name=column_name,
                            value=try_float(row[column_name], column_name),
                        )
                        for column_name in normal_columns
                    ],
                )
                for _, row in data.iterrows()
            ],
            aggregated_entries=[
                GroupDataElementEntry(
                    name=column_name,
                    value=try_float(top_row[column_name], column_name),
                )
                for column_name in aggregated_columns
            ],
        )


@dataclass(frozen=True)
class GroupData:
    name: str
    sample_data: list[GroupSampleData]

    @staticmethod
    def create(reader: CsvReader) -> GroupData:
        name = assert_not_none(
            reader.pop(),
            msg="Unable to find group block name.",
        ).removeprefix("Group: ")

        data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t", header=0),
            msg="Unable to find group block data.",
        ).replace(r"^\s+$", None, regex=True)

        assert_not_none(
            data.get("Sample"),
            msg=f"Unable to find sample identifier column in group data {name}",
        )

        samples = data["Sample"].ffill()
        return GroupData(
            name=name,
            sample_data=[
                GroupSampleData.create(data.iloc[sample_entries.index])
                for _, sample_entries in samples.groupby(samples)
            ],
        )


@dataclass(frozen=True)
class GroupColumns:
    data: dict[str, str]

    @staticmethod
    def create(reader: CsvReader) -> GroupColumns:
        data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t", header=0),
            msg="Unable to find group block columns.",
        )

        if "Formula Name" not in data:
            error = "Unable to find formula name in group block columns."
            raise AllotropeConversionError(error)

        if "Formula" not in data:
            error = "Unable to find formula in group block columns."
            raise AllotropeConversionError(error)

        return GroupColumns(
            data=dict(zip(data["Formula Name"], data["Formula"])),
        )


@dataclass(frozen=True)
class GroupSummaries:
    data: list[str]

    @staticmethod
    def create(reader: CsvReader) -> GroupSummaries:
        data = list(reader.pop_until_empty())
        reader.drop_empty()
        return GroupSummaries(data)


@dataclass(frozen=True)
class GroupBlock(Block):
    group_data: GroupData
    group_columns: GroupColumns
    group_summaries: GroupSummaries

    @staticmethod
    def create(reader: CsvReader) -> GroupBlock:
        return GroupBlock(
            block_type="Group",
            group_data=GroupData.create(reader),
            group_columns=GroupColumns.create(reader),
            group_summaries=GroupSummaries.create(reader),
        )


# TODO do we need to do anything with these?
@dataclass(frozen=True)
class NoteBlock(Block):
    pass


@dataclass(frozen=True)
class PlateHeader:
    name: str
    export_version: str
    export_format: str
    read_type: str
    data_type: str
    kinetic_points: int
    num_wavelengths: int
    wavelengths: list[float]
    num_columns: int
    num_wells: int
    concept: str
    read_mode: str
    unit: str
    scan_position: ScanPosition
    reads_per_well: float
    pmt_gain: Optional[str]
    num_rows: int
    excitation_wavelengths: Optional[list[int]]
    cutoff_filters: Optional[list[int]]


@dataclass
class DataElement:
    uuid: str
    plate: str
    temperature: Optional[float]
    wavelength: float
    position: str
    value: JsonFloat
    sample_id: Optional[str] = None

    @property
    def sample_identifier(self) -> str:
        return self.sample_id if self.sample_id else f"{self.plate} {self.position}"


@dataclass(frozen=True)
class ReducedDataElement:
    position: str
    value: float


@dataclass(frozen=True)
class PlateWavelengthData:
    wavelength: float
    data_elements: dict[str, DataElement]

    @staticmethod
    def create(
        plate_name: str,
        temperature: Optional[float],
        wavelength: float,
        df_data: pd.DataFrame,
    ) -> PlateWavelengthData:
        data = {
            f"{num_to_chars(row_idx)}{col}": value
            for row_idx, *row_data in df_data.itertuples()
            for col, value in zip(df_data.columns, row_data)
        }
        return PlateWavelengthData(
            wavelength,
            data_elements={
                str(position): DataElement(
                    uuid=random_uuid_str(),
                    plate=plate_name,
                    temperature=temperature,
                    wavelength=wavelength,
                    position=str(position),
                    value=try_float_or_nan(value),
                )
                for position, value in data.items()
            },
        )


@dataclass(frozen=True)
class PlateKineticData:
    temperature: Optional[float]
    wavelength_data: list[PlateWavelengthData]

    @staticmethod
    def create(
        reader: CsvReader,
        header: PlateHeader,
        columns: pd.Series[str],
    ) -> PlateKineticData:

        # use plate dimensions to determine how many rows of plate block to read
        dimensions = assert_not_none(
            NUM_WELLS_TO_PLATE_DIMENSIONS.get(header.num_wells),
            msg="unable to determine plate dimensions",
        )
        rows = dimensions[1]
        lines = []
        # read number of rows in plate
        for _row in range(rows):
            lines.append(reader.pop() or "")
        reader.drop_empty()

        # convert rows to df
        data = assert_not_none(
            reader.lines_as_df(lines=lines, sep="\t"),
            msg="unable to find data from plate block.",
        )
        data.columns = pd.Index(columns)

        # get temprature from the first column of the first row with value
        temperature = try_float_or_none(
            str(data.iloc[int(pd.to_numeric(data.first_valid_index())), 1])
        )

        return PlateKineticData(
            temperature=temperature,
            wavelength_data=PlateKineticData._get_wavelength_data(
                header.name,
                temperature,
                header,
                data.iloc[:, 2:],
            ),
        )

    @staticmethod
    def _get_wavelength_data(
        plate_name: str,
        temperature: Optional[float],
        header: PlateHeader,
        w_data: pd.DataFrame,
    ) -> list[PlateWavelengthData]:
        wavelength_data = []
        for idx in range(header.num_wavelengths):
            wavelength = header.wavelengths[idx]
            start = idx * (header.num_columns + 1)
            end = start + header.num_columns
            wavelength_data.append(
                PlateWavelengthData.create(
                    plate_name,
                    temperature,
                    wavelength,
                    w_data.iloc[:, start:end],
                )
            )
        return wavelength_data


@dataclass(frozen=True)
class PlateRawData:
    kinetic_data: list[PlateKineticData]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> PlateRawData:
        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find data columns for plate block raw data.",
        )

        return PlateRawData(
            kinetic_data=[
                PlateKineticData.create(reader, header, columns)
                for _ in range(header.kinetic_points)
            ]
        )


@dataclass(frozen=True)
class PlateReducedData:
    data: list[ReducedDataElement]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> PlateReducedData:
        raw_data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t", header=0),
            msg="Unable to find reduced data for plate block.",
        )

        start = 2
        df_data = raw_data.iloc[:, start : (start + header.num_columns)]

        reduced_data_elements = []
        for row, *data in df_data.itertuples():
            for col, str_value in zip(df_data.columns, data):
                value = try_non_nan_float_or_none(str_value)
                if value is not None:
                    reduced_data_elements.append(
                        ReducedDataElement(
                            position=f"{num_to_chars(row)}{col}",
                            value=value,
                        )
                    )
        return PlateReducedData(data=reduced_data_elements)


@dataclass(frozen=True)
class PlateData:
    raw_data: PlateRawData
    reduced_data: Optional[PlateReducedData]

    @staticmethod
    def create(
        reader: CsvReader,
        header: PlateHeader,
    ) -> PlateData:
        return PlateData(
            raw_data=PlateRawData.create(reader, header),
            reduced_data=(
                PlateReducedData.create(reader, header)
                if reader.current_line_exists()
                else None
            ),
        )

    def iter_data_elements(self, position: str) -> Iterator[DataElement]:
        for kinetic_data in self.raw_data.kinetic_data:
            for wavelength_data in kinetic_data.wavelength_data:
                yield wavelength_data.data_elements[position]


@dataclass(frozen=True)
class TimeKineticData:
    temperature: Optional[float]
    data_elements: dict[str, DataElement]

    @staticmethod
    def create(
        plate_name: str,
        wavelength: float,
        row: pd.Series[float],
    ) -> TimeKineticData:
        temperature = try_non_nan_float_or_none(str(row.iloc[1]))

        return TimeKineticData(
            temperature=temperature,
            data_elements={
                str(position): DataElement(
                    uuid=random_uuid_str(),
                    plate=plate_name,
                    temperature=temperature,
                    wavelength=wavelength,
                    position=str(position),
                    value=try_float_or_nan(str(value)),
                )
                for position, value in row.iloc[2:].items()
            },
        )


@dataclass(frozen=True)
class TimeWavelengthData:
    wavelength: float
    kinetic_data: list[TimeKineticData]

    @staticmethod
    def create(
        reader: CsvReader,
        plate_name: str,
        wavelength: float,
        columns: pd.Series[str],
    ) -> TimeWavelengthData:
        data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t"),
            msg="unable to find raw data from time block.",
        )
        data.columns = pd.Index(columns)
        return TimeWavelengthData(
            wavelength=wavelength,
            kinetic_data=[
                TimeKineticData.create(plate_name, wavelength, row)
                for _, row in data.iterrows()
            ],
        )


@dataclass(frozen=True)
class TimeRawData:
    wavelength_data: list[TimeWavelengthData]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> TimeRawData:
        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find data columns for time block raw data.",
        )

        return TimeRawData(
            wavelength_data=[
                TimeWavelengthData.create(
                    reader,
                    header.name,
                    wavelength,
                    columns,
                )
                for wavelength in header.wavelengths
            ]
        )


@dataclass(frozen=True)
class TimeReducedData:
    data: list[ReducedDataElement]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> TimeReducedData:
        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find columns for time block reduced data.",
        )
        data = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find reduced data from time block.",
        )
        data.index = pd.Index(columns)

        reduced_data_elements = []
        for pos, str_value in data[2 : header.num_wells + 2].items():
            value = try_non_nan_float_or_none(str_value)
            if value is not None:
                reduced_data_elements.append(
                    ReducedDataElement(
                        position=str(pos),
                        value=value,
                    )
                )
        return TimeReducedData(data=reduced_data_elements)


@dataclass(frozen=True)
class TimeData:
    raw_data: TimeRawData
    reduced_data: Optional[TimeReducedData]

    @staticmethod
    def create(
        reader: CsvReader,
        header: PlateHeader,
    ) -> TimeData:
        return TimeData(
            raw_data=TimeRawData.create(reader, header),
            reduced_data=(
                TimeReducedData.create(reader, header)
                if reader.current_line_exists()
                else None
            ),
        )

    def iter_data_elements(self, position: str) -> Iterator[DataElement]:
        for wavelength_data in self.raw_data.wavelength_data:
            for kinetic_data in wavelength_data.kinetic_data:
                yield kinetic_data.data_elements[position]


@dataclass(frozen=True)
class PlateBlock(ABC, Block):
    header: PlateHeader
    block_data: Union[PlateData, TimeData]

    @staticmethod
    def read_header(reader: CsvReader) -> pd.Series[str]:
        raw_header_series = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="Unable to find plate block header.",
        )
        return raw_header_series.replace("", None).str.strip()

    @staticmethod
    def get_plate_block_cls(header_series: pd.Series[str]) -> type[PlateBlock]:
        plate_block_cls: dict[str, type[PlateBlock]] = {
            "Absorbance": AbsorbancePlateBlock,
            "Fluorescence": FluorescencePlateBlock,
            "Luminescence": LuminescencePlateBlock,
        }
        read_mode = header_series[5]
        cls = plate_block_cls.get(read_mode or "")
        if cls is None:
            msg = msg_for_error_on_unrecognized_value(
                "read mode", read_mode, plate_block_cls.keys()
            )
            raise AllotropeConversionError(msg)
        return cls

    @staticmethod
    @abstractmethod
    def get_plate_block_type() -> str:
        ...

    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        raise NotImplementedError

    @classmethod
    def check_export_version(cls, export_version: str) -> None:
        if export_version != EXPORT_VERSION:
            error = f"Unsupported export version {export_version}; only {EXPORT_VERSION} is supported."
            raise AllotropeConversionError(error)

    @classmethod
    def check_read_type(cls, read_type: str) -> None:
        if read_type != ReadType.ENDPOINT.value:
            error = "Only Endpoint measurements can be processed at this time."
            raise AllotropeConversionError(error)

    @classmethod
    def check_data_type(cls, data_type: str) -> None:
        if data_type not in (DataType.RAW.value, DataType.BOTH.value):
            error = "The SoftMax Pro file is required to include either 'Raw' or 'Both' (Raw and Reduced) data for all plates"
            raise AllotropeConversionError(error)

    @classmethod
    def check_num_wavelengths(
        cls, wavelengths: list[float], num_wavelengths: int
    ) -> None:
        if len(wavelengths) != num_wavelengths:
            error = "Unable to find expected number of wavelength values."
            raise AllotropeConversionError(error)

    @classmethod
    def get_num_wavelengths(cls, num_wavelengths_raw: Optional[str]) -> int:
        return try_int_or_none(num_wavelengths_raw) or 1

    @classmethod
    def get_wavelengths(cls, wavelengths_str: Optional[str]) -> list[float]:
        return [
            try_float(wavelength, "wavelength")
            for wavelength in assert_not_none(
                wavelengths_str,
                msg="Unable to find wavelengths list.",
            ).split()
        ]

    def iter_wells(self) -> Iterator[str]:
        cols, rows = NUM_WELLS_TO_PLATE_DIMENSIONS[self.header.num_wells]
        for row in range(rows):
            for col in range(1, cols + 1):
                yield f"{num_to_chars(row)}{col}"

    def iter_data_elements(self, position: str) -> Iterator[DataElement]:
        yield from self.block_data.iter_data_elements(position)

    def iter_reduced_data(self) -> Iterator[ReducedDataElement]:
        if self.block_data.reduced_data:
            yield from self.block_data.reduced_data.data


@dataclass(frozen=True)
class FluorescencePlateBlock(PlateBlock):
    @staticmethod
    def get_plate_block_type() -> str:
        return "Fluorescence"

    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            _,  # Read mode
            raw_scan_position,
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            _,  # read_time_or_scan_pattern
            _,  # read_interval_or_scan_density
            _,  # start_wavelength
            _,  # end_wavelength
            _,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            _,  # num_columns
            num_wells_raw,
            excitation_wavelengths_str,
            _,  # cutoff
            cutoff_filters_str,
            _,  # sweep_wave
            _,  # sweep_wavelength
            reads_per_well,
            pmt_gain,
            _,  # start_integration_time
            _,  # end_integration_time
            _,  # first_row
            num_rows,
        ] = header[:31]

        cls.check_export_version(export_version)
        cls.check_read_type(read_type)
        cls.check_data_type(data_type)

        num_wavelengths = cls.get_num_wavelengths(num_wavelengths_raw)
        wavelengths = cls.get_wavelengths(wavelengths_str)
        cls.check_num_wavelengths(wavelengths, num_wavelengths)

        assert_not_none(
            excitation_wavelengths_str,
            msg="Unable to find excitation wavelengths.",
        )

        excitation_wavelengths = [
            try_int(excitation_wavelength, "excitation wavelengths")
            for excitation_wavelength in excitation_wavelengths_str.split()
        ]

        if len(excitation_wavelengths) != num_wavelengths:
            error = "Unable to find expected number of excitation values."
            raise AllotropeConversionError(error)

        # cutoff filters is an optional field in the input file
        # if present it contains a list of string numbers separated by spaces
        cutoff_filters = (
            [
                try_int(cutoff_filters, "cutoff filters")
                for cutoff_filters in cutoff_filters_str.split()
            ]
            if cutoff_filters_str is not None
            else None
        )

        # if there are cutoff filters check that the size match number of wavelengths
        if cutoff_filters is not None and len(cutoff_filters) != num_wavelengths:
            error = "Unable to find expected number of cutoff filter values."
            raise AllotropeConversionError(error)

        if raw_scan_position == "TRUE":
            scan_position = ScanPosition.BOTTOM
        elif raw_scan_position == "FALSE":
            scan_position = ScanPosition.TOP
        else:
            error = f"{raw_scan_position} is not a valid scan position."
            raise AllotropeConversionError(error)

        num_wells = try_int(num_wells_raw, "num_wells")
        num_columns = num_wells_to_n_columns(num_wells)

        return PlateHeader(
            name=name,
            export_version=export_version,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=try_int(kinetic_points_raw, "kinetic_points"),
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            concept="fluorescence",
            read_mode="Fluorescence",
            unit="RFU",
            scan_position=scan_position,
            reads_per_well=try_float(reads_per_well, "reads_per_well"),
            pmt_gain=pmt_gain,
            num_rows=try_int(num_rows, "num_rows"),
            excitation_wavelengths=excitation_wavelengths,
            cutoff_filters=cutoff_filters,
        )


@dataclass(frozen=True)
class LuminescencePlateBlock(PlateBlock):
    @staticmethod
    def get_plate_block_type() -> str:
        return "Luminescence"

    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            _,  # Read mode
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            _,  # read_time_or_scan_pattern
            _,  # read_interval_or_scan_density
            _,  # start_wavelength
            _,  # end_wavelength
            _,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            _,  # num_columns,
            num_wells_raw,
            _,  # excitation_wavelengths_str
            _,  # cutoff
            _,  # cutoff_filters_str,
            _,  # sweep_wave
            _,  # sweep_wavelength
            reads_per_well,
            pmt_gain,
            _,  # start_integration_time
            _,  # end_integration_time
            _,  # first_row
            num_rows,
        ] = header[:30]

        cls.check_export_version(export_version)
        cls.check_read_type(read_type)
        cls.check_data_type(data_type)

        num_wavelengths = cls.get_num_wavelengths(num_wavelengths_raw)
        wavelengths = cls.get_wavelengths(wavelengths_str)
        cls.check_num_wavelengths(wavelengths, num_wavelengths)

        num_wells = try_int(num_wells_raw, "num_wells")
        num_columns = num_wells_to_n_columns(num_wells)

        return PlateHeader(
            name=name,
            export_version=export_version,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=try_int(kinetic_points_raw, "kinetic_points"),
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            concept="luminescence",
            read_mode="Luminescence",
            unit="RLU",
            scan_position=ScanPosition.NONE,
            reads_per_well=try_int(reads_per_well, "reads_per_well"),
            pmt_gain=pmt_gain,
            num_rows=try_int(num_rows, "num_rows"),
            excitation_wavelengths=None,
            cutoff_filters=None,
        )


@dataclass(frozen=True)
class AbsorbancePlateBlock(PlateBlock):
    @staticmethod
    def get_plate_block_type() -> str:
        return "Absorbance"

    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            _,  # Read mode
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            _,  # read_time_or_scan_pattern
            _,  # read_interval_or_scan_density
            _,  # start_wavelength
            _,  # end_wavelength
            _,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            _,  # num_columns
            num_wells_raw,
            _,
            num_rows_raw,
        ] = header[:21]

        cls.check_export_version(export_version)
        cls.check_read_type(read_type)
        cls.check_data_type(data_type)

        num_wavelengths = cls.get_num_wavelengths(num_wavelengths_raw)
        wavelengths = cls.get_wavelengths(wavelengths_str)
        cls.check_num_wavelengths(wavelengths, num_wavelengths)

        num_wells = try_int(num_wells_raw, "num_wells")
        num_columns = num_wells_to_n_columns(num_wells)

        return PlateHeader(
            name=name,
            export_version=export_version,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=try_int(kinetic_points_raw, "kinetic_points"),
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            concept="absorbance",
            read_mode="Absorbance",
            unit="mAU",
            scan_position=ScanPosition.NONE,
            reads_per_well=0,
            pmt_gain=None,
            num_rows=try_int(num_rows_raw, "num_rows"),
            excitation_wavelengths=None,
            cutoff_filters=None,
        )


@dataclass(frozen=True)
class BlockList:
    plate_blocks: dict[str, PlateBlock]
    group_blocks: list[GroupBlock]

    @staticmethod
    def create(reader: CsvReader) -> BlockList:
        plate_blocks = {}
        group_blocks = []

        for sub_reader in BlockList._iter_blocks_reader(reader):
            if sub_reader.match("^Group"):
                if "WellPlateName" in assert_not_none(
                    sub_reader.get_line(sub_reader.current_line + 1),
                    msg="Unable to get columns from group block",
                ):
                    group_blocks.append(GroupBlock.create(sub_reader))
            elif sub_reader.match("^Plate"):
                header_series = PlateBlock.read_header(sub_reader)
                cls = PlateBlock.get_plate_block_cls(header_series)
                header = cls.parse_header(header_series)

                block_data: Union[TimeData, PlateData]
                if header.export_format == ExportFormat.TIME_FORMAT.value:
                    block_data = TimeData.create(sub_reader, header)
                elif header.export_format == ExportFormat.PLATE_FORMAT.value:
                    block_data = PlateData.create(sub_reader, header)
                else:
                    error = f"unrecognized export format {header.export_format}"
                    raise AllotropeConversionError(error)

                plate_blocks[header.name] = cls(
                    block_type="Plate",
                    header=header,
                    block_data=block_data,
                )
            elif not sub_reader.match("^Note"):
                error = f"Expected block '{sub_reader.get()}' to start with Group, Plate or Note."
                raise AllotropeConversionError(error)

        return BlockList(
            plate_blocks=plate_blocks,
            group_blocks=group_blocks,
        )

    @staticmethod
    def _get_n_blocks(reader: CsvReader) -> int:
        start_line = reader.pop() or ""
        if search_result := re.search(BLOCKS_LINE_REGEX, start_line):
            return int(search_result.group(1))
        msg = msg_for_error_on_unrecognized_value("start line", start_line)
        raise AllotropeConversionError(msg)

    @staticmethod
    def _iter_blocks_reader(reader: CsvReader) -> Iterator[CsvReader]:
        n_blocks = BlockList._get_n_blocks(reader)
        for _ in range(n_blocks):
            yield CsvReader(list(reader.pop_until(END_LINE_REGEX)))
            reader.pop()  # drop end line
            reader.drop_empty()


@dataclass(frozen=True)
class Data:
    block_list: BlockList

    @staticmethod
    def create(reader: CsvReader) -> Data:
        block_list = BlockList.create(reader)

        for group_block in block_list.group_blocks:
            for group_sample_data in group_block.group_data.sample_data:
                for group_data_element in group_sample_data.data_elements:
                    plate_block = block_list.plate_blocks[group_data_element.plate]
                    for data_element in plate_block.iter_data_elements(
                        group_data_element.position
                    ):
                        data_element.sample_id = group_data_element.sample

        return Data(block_list)



from io import StringIO

import pandas as pd

from allotropy.allotrope.pandas_util import read_csv
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers import lines_reader
from allotropy.parsers.lines_reader import CsvReader


class NanoDropEightReader:
    @classmethod
    def read(cls, named_file_contents: NamedFileContents) -> pd.DataFrame:
        all_lines = lines_reader.read_to_lines(named_file_contents)
        reader = CsvReader(all_lines)
        lines = reader.pop_csv_block_as_lines()
        raw_data = read_csv(
            StringIO("\n".join(lines)),
            sep="\t",
            dtype={"Plate ID": str, "Sample ID": str},
            # Prevent pandas from rounding decimal values, at the cost of some speed.
            float_precision="round_trip",
        )
        raw_data = raw_data.rename(columns=lambda x: x.strip().lower())

        return raw_data



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.unchained_labs_lunatic.constants import (
    CALCULATED_DATA_LOOKUP,
    INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG,
    NO_DATE_OR_TIME_ERROR_MSG,
    NO_MEASUREMENT_IN_PLATE_ERROR_MSG,
    NO_WAVELENGTH_COLUMN_ERROR_MSG,
    WAVELENGTH_COLUMNS_RE,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    try_float_from_series_or_nan,
    try_float_from_series_or_none,
    try_str_from_series,
    try_str_from_series_or_none,
)


@dataclass(frozen=True)
class DataSourceItem:
    identifier: str
    feature: str


@dataclass(frozen=True)
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_source_document: list[DataSourceItem]


@dataclass(frozen=True)
class Measurement:
    identifier: str
    wavelength: float
    absorbance: JsonFloat
    sample_identifier: str
    location_identifier: str
    well_plate_identifier: Optional[str]
    calculated_data: list[CalculatedDataItem]

    @staticmethod
    def create(well_plate_data: pd.Series[Any], wavelength_column: str) -> Measurement:
        if wavelength_column not in well_plate_data:
            msg = NO_MEASUREMENT_IN_PLATE_ERROR_MSG.format(wavelength_column)
            raise AllotropeConversionError(msg)

        if not WAVELENGTH_COLUMNS_RE.match(wavelength_column):
            raise AllotropeConversionError(INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG)

        measurement_identifier = random_uuid_str()
        return Measurement(
            identifier=measurement_identifier,
            wavelength=float(wavelength_column[1:]),
            absorbance=try_float_from_series_or_nan(well_plate_data, wavelength_column),
            sample_identifier=try_str_from_series(well_plate_data, "Sample name"),
            location_identifier=try_str_from_series(well_plate_data, "Plate Position"),
            well_plate_identifier=try_str_from_series_or_none(
                well_plate_data, "Plate ID"
            ),
            calculated_data=Measurement._get_calculated_data(
                well_plate_data, wavelength_column, measurement_identifier
            ),
        )

    @staticmethod
    def _get_calculated_data(
        well_plate_data: pd.Series[Any],
        wavelength_column: str,
        measurement_identifier: str,
    ) -> list[CalculatedDataItem]:
        calculated_data_dict = CALCULATED_DATA_LOOKUP.get(wavelength_column)
        if not calculated_data_dict:
            return []

        calculated_data = []
        for item in calculated_data_dict:
            value = try_float_from_series_or_none(well_plate_data, item["column"])
            if value is None:
                continue

            calculated_data.append(
                CalculatedDataItem(
                    identifier=random_uuid_str(),
                    name=item["name"],
                    value=value,
                    unit=item["unit"],
                    data_source_document=[
                        DataSourceItem(
                            identifier=measurement_identifier,
                            feature=item["feature"],
                        )
                    ],
                )
            )
        return calculated_data


@dataclass(frozen=True)
class WellPlate:
    measurement_time: str
    analytical_method_identifier: Optional[str]
    measurements: list[Measurement]

    @staticmethod
    def create(plate_data: pd.Series[Any], wavelength_columns: list[str]) -> WellPlate:
        return WellPlate(
            measurement_time=WellPlate._get_datetime_from_plate(plate_data),
            analytical_method_identifier=try_str_from_series_or_none(
                plate_data, "Application"
            ),
            measurements=[
                Measurement.create(plate_data, wavelength_column)
                for wavelength_column in wavelength_columns
            ],
        )

    @staticmethod
    def _get_datetime_from_plate(plate_data: pd.Series[Any]) -> str:
        date = plate_data.get("Date")
        time = plate_data.get("Time")

        if not date or not time:
            raise AllotropeConversionError(NO_DATE_OR_TIME_ERROR_MSG)

        return f"{date} {time}"


@dataclass(frozen=True)
class Data:
    device_identifier: str
    well_plate_list: list[WellPlate]

    @staticmethod
    def create(data: pd.DataFrame) -> Data:
        device_identifier = Data._get_device_identifier(data.iloc[0])

        wavelength_columns = list(filter(WAVELENGTH_COLUMNS_RE.match, data.columns))
        if not wavelength_columns:
            raise AllotropeConversionError(NO_WAVELENGTH_COLUMN_ERROR_MSG)

        return Data(
            device_identifier=device_identifier,
            well_plate_list=[
                WellPlate.create(data.iloc[i], wavelength_columns)
                for i in range(len(data.index))
            ],
        )

    @staticmethod
    def _get_device_identifier(data: pd.Series[Any]) -> str:
        device_identifier = try_str_from_series(data, "Instrument ID")

        return str(device_identifier)

    def get_calculated_data_document(self) -> list[CalculatedDataItem]:
        calculated_data_document = []
        for well_plate in self.well_plate_list:
            for measurement in well_plate.measurements:
                calculated_data_document.extend(measurement.calculated_data)

        return calculated_data_document



from __future__ import annotations

from dataclasses import dataclass
from io import StringIO
import re
from typing import Any, Optional

import pandas as pd

from allotropy.allotrope.pandas_util import read_csv
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
    try_float_from_series,
    try_str_from_series,
    try_str_from_series_or_none,
)

LUMINEX_EMPTY_PATTERN = r"^[,\"\s]*$"
CALIBRATION_BLOCK_HEADER = "Most Recent Calibration and Verification Results"
TABLE_HEADER_PATTERN = '"DataType:","{}"'
MINIMUM_CALIBRATION_LINE_COLS = 2
EXPECTED_CALIBRATION_RESULT_LEN = 2


@dataclass(frozen=True)
class Header:
    model_number: str
    software_version: str
    equipment_serial_number: str
    analytical_method_identifier: str
    method_version: str
    experimental_data_identifier: str
    sample_volume_setting: float
    plate_well_count: float
    measurement_time: str
    detector_gain_setting: str
    data_system_instance_identifier: str
    analyst: Optional[str] = None

    @classmethod
    def create(cls, header_data: pd.DataFrame) -> Header:
        info_row = header_data.iloc[0]
        raw_datetime = try_str_from_series(info_row, "BatchStartTime")

        return Header(
            model_number=cls._get_model_number(header_data),
            software_version=try_str_from_series(info_row, "Build"),
            equipment_serial_number=try_str_from_series(info_row, "SN"),
            analytical_method_identifier=try_str_from_series(info_row, "ProtocolName"),
            method_version=try_str_from_series(info_row, "ProtocolVersion"),
            experimental_data_identifier=try_str_from_series(info_row, "Batch"),
            sample_volume_setting=cls._get_sample_volume_setting(info_row),
            plate_well_count=cls._get_plate_well_count(header_data),
            measurement_time=raw_datetime,
            detector_gain_setting=try_str_from_series(info_row, "ProtocolReporterGain"),
            data_system_instance_identifier=try_str_from_series(
                info_row, "ComputerName"
            ),
            analyst=try_str_from_series_or_none(info_row, "Operator"),
        )

    @classmethod
    def _get_sample_volume_setting(cls, info_row: pd.Series[Any]) -> float:
        sample_volume = try_str_from_series(info_row, "SampleVolume")

        return try_float(sample_volume.split()[0], "sample volume setting")

    @classmethod
    def _get_model_number(cls, header_data: pd.DataFrame) -> str:
        program_data = cls._try_col_from_header(header_data, "Program")

        try:
            model_number = program_data.iloc[2]
        except IndexError as e:
            msg = "Unable to find model number in Program row."
            raise AllotropeConversionError(msg) from e

        return str(model_number)

    @classmethod
    def _get_plate_well_count(cls, header_data: pd.DataFrame) -> float:
        protocol_plate_data = cls._try_col_from_header(header_data, "ProtocolPlate")

        try:
            plate_well_count = protocol_plate_data.iloc[3]
        except IndexError as e:
            msg = "Unable to find plate well count in ProtocolPlate row."
            raise AllotropeConversionError(msg) from e

        return try_float(plate_well_count, "plate well count")

    @classmethod
    def _try_col_from_header(
        cls, header_data: pd.DataFrame, key: str
    ) -> pd.Series[Any]:
        if key not in header_data:
            msg = f"Unable to find {key} data on header block."
            raise AllotropeConversionError(msg)

        return header_data[key]


@dataclass(frozen=True)
class CalibrationItem:
    name: str
    report: str
    time: str

    @classmethod
    def create(cls, calibration_line: str) -> CalibrationItem:
        """Create a CalibrationItem from a calibration line.

        Each line should follow the pattern "Last <calibration_name>,<calibration_report> <calibration_time><,,,,"
        example: "Last F3DeCAL1 Calibration,Passed 05/17/2023 09:25:11,,,,,,"
        """
        calibration_data = calibration_line.replace('"', "").split(",")
        if len(calibration_data) < MINIMUM_CALIBRATION_LINE_COLS:
            msg = f"Expected at least two columns on the calibration line, got: {calibration_line}"
            raise AllotropeConversionError(msg)

        calibration_result = calibration_data[1].split(maxsplit=1)

        if len(calibration_result) != EXPECTED_CALIBRATION_RESULT_LEN:
            msg = f"Invalid calibration result format, got: {calibration_data[1]}"
            raise AllotropeConversionError(msg)

        return CalibrationItem(
            name=calibration_data[0].replace("Last", "").strip(),
            report=calibration_result[0],
            time=calibration_result[1],
        )


@dataclass(frozen=True)
class Analyte:
    analyte_name: str
    assay_bead_identifier: str
    assay_bead_count: float
    fluorescence: float


@dataclass(frozen=True)
class Measurement:
    sample_identifier: str
    location_identifier: str
    dilution_factor_setting: float
    assay_bead_count: float
    analytes: list[Analyte]
    errors: Optional[list[str]] = None

    @classmethod
    def create(
        cls,
        median_data: pd.Series[Any],
        count_data: pd.DataFrame,
        bead_ids_data: pd.Series[str],
        dilution_factor_data: pd.DataFrame,
        errors_data: pd.DataFrame,
    ) -> Measurement:
        location = try_str_from_series(median_data, "Location")
        dilution_factor_setting = try_float_from_series(
            dilution_factor_data.loc[location], "Dilution Factor"
        )
        # analyte names are columns 3 through the penultimate
        analyte_names = list(median_data.index)[2:-1]

        well_location, location_id = cls._get_location_details(location)

        return Measurement(
            sample_identifier=try_str_from_series(median_data, "Sample"),
            location_identifier=location_id,
            dilution_factor_setting=dilution_factor_setting,
            assay_bead_count=try_float_from_series(median_data, "Total Events"),
            analytes=[
                Analyte(
                    analyte_name=analyte,
                    assay_bead_identifier=try_str_from_series(bead_ids_data, analyte),
                    assay_bead_count=try_float_from_series(
                        count_data.loc[location], analyte
                    ),
                    fluorescence=try_float_from_series(median_data, analyte),
                )
                for analyte in analyte_names
            ],
            errors=cls._get_errors(errors_data, well_location),
        )

    @classmethod
    def _get_location_details(cls, location: str) -> tuple[str, str]:
        location_regex = r"\d+\((?P<well_location>\d+,(?P<location_id>\w+))\)"
        match = assert_not_none(
            re.search(location_regex, location),
            msg=f"Invalid location format: {location}",
        )

        return match.group("well_location"), match.group("location_id")

    @classmethod
    def _get_errors(
        cls, errors_data: pd.DataFrame, well_location: str
    ) -> Optional[list[str]]:
        try:
            measurement_errors = errors_data.loc[well_location]
        except KeyError:
            return None

        return [
            measurement_errors.iloc[i]["Message"]
            for i in range(len(measurement_errors))
        ]


@dataclass(frozen=True)
class MeasurementList:
    measurements: list[Measurement]

    @classmethod
    def create(cls, reader: CsvReader) -> MeasurementList:
        median_data = cls._get_median_data(reader)

        count_data = cls._get_table_as_df(reader, "Count")
        bead_ids_data = cls._get_bead_ids_data(reader)
        dilution_factor_data = cls._get_table_as_df(reader, "Dilution Factor")
        errors_data = cls._get_table_as_df(reader, "Warnings/Errors")

        return MeasurementList(
            measurements=[
                Measurement.create(
                    median_data=median_data.iloc[i],
                    count_data=count_data,
                    bead_ids_data=bead_ids_data,
                    dilution_factor_data=dilution_factor_data,
                    errors_data=errors_data,
                )
                for i in range(len(median_data))
            ]
        )

    @classmethod
    def _get_median_data(cls, reader: CsvReader) -> pd.DataFrame:
        reader.drop_until_inclusive(TABLE_HEADER_PATTERN.format("Median"))
        return assert_not_none(
            reader.pop_csv_block_as_df(empty_pat=LUMINEX_EMPTY_PATTERN, header="infer"),
            msg="Unable to find Median table.",
        )

    @classmethod
    def _get_bead_ids_data(cls, reader: CsvReader) -> pd.Series[str]:
        units_df = MeasurementList._get_table_as_df(reader, "Units")
        return units_df.loc["BeadID:"]

    @classmethod
    def _get_table_as_df(cls, reader: CsvReader, table_name: str) -> pd.DataFrame:
        """Returns a dataframe that has the well location as index.

        Results tables in luminex xponent output files have the location as first column.
        Having this column as the index of the dataframe allows for easier lookup when
        retrieving measurement data.
        """
        reader.drop_until_inclusive(match_pat=TABLE_HEADER_PATTERN.format(table_name))

        table_lines = reader.pop_csv_block_as_lines(empty_pat=LUMINEX_EMPTY_PATTERN)

        if not table_lines:
            msg = f"Unable to find {table_name} table."
            raise AllotropeConversionError(msg)

        return read_csv(
            StringIO("\n".join(table_lines)),
            header=[0],
            index_col=[0],
        ).dropna(how="all", axis=1)


@dataclass(frozen=True)
class Data:
    header: Header
    calibration_data: list[CalibrationItem]
    minimum_bead_count_setting: float
    measurement_list: MeasurementList

    @classmethod
    def create(cls, reader: CsvReader) -> Data:
        return Data(
            header=Header.create(cls._get_header_data(reader)),
            calibration_data=cls._get_calibration_data(reader),
            minimum_bead_count_setting=cls._get_minimum_bead_count_setting(reader),
            measurement_list=MeasurementList.create(reader),
        )

    @classmethod
    def _get_header_data(cls, reader: CsvReader) -> pd.DataFrame:
        header_lines = assert_not_none(
            reader.pop_until(CALIBRATION_BLOCK_HEADER), "Unable to find Header block."
        )

        header_data = read_csv(
            StringIO("\n".join(header_lines)),
            header=None,
            index_col=0,
            names=range(7),
        ).dropna(how="all")

        return header_data.T

    @classmethod
    def _get_calibration_data(cls, reader: CsvReader) -> list[CalibrationItem]:
        reader.drop_until_inclusive(CALIBRATION_BLOCK_HEADER)
        calibration_lines = reader.pop_csv_block_as_lines(
            empty_pat=LUMINEX_EMPTY_PATTERN
        )
        if not calibration_lines:
            msg = "Unable to find Calibration Block."
            raise AllotropeConversionError(msg)

        calibration_list = []

        for line in calibration_lines:
            calibration_list.append(CalibrationItem.create(line))

        return calibration_list

    @classmethod
    def _get_minimum_bead_count_setting(cls, reader: CsvReader) -> float:
        reader.drop_until(match_pat='"Samples",')
        samples_info = assert_not_none(reader.pop(), "Unable to find Samples info.")
        try:
            min_bead_count_setting = samples_info.replace('"', "").split(",")[3]
        except IndexError as e:
            msg = "Unable to find minimum bead count setting in Samples info."
            raise AllotropeConversionError(msg) from e

        return try_float(min_bead_count_setting, "minimum bead count setting")



from __future__ import annotations

from dataclasses import dataclass
import re
from typing import Any, Optional

import numpy as np
import pandas as pd

from allotropy.allotrope.pandas_util import read_csv
from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.novabio_flex2.constants import (
    ANALYTE_MAPPINGS,
    FILENAME_REGEX,
    INVALID_FILENAME_MESSAGE,
    MOLAR_CONCENTRATION_CLASSES,
    MOLAR_CONCENTRATION_CLS_BY_UNIT,
    PROPERTY_MAPPINGS,
)


@dataclass(frozen=True)
class Title:
    processing_time: str
    device_identifier: Optional[str]

    @staticmethod
    def create(filename: str) -> Title:
        matches = re.match(FILENAME_REGEX, filename)

        if not matches:
            raise AllotropeConversionError(INVALID_FILENAME_MESSAGE.format(filename))

        matches_dict = matches.groupdict()
        return Title(
            processing_time=matches_dict["processing_time"].replace("_", " "),
            device_identifier=matches_dict["device_identifier"] or None,
        )


@dataclass(frozen=True)
class Analyte:
    name: str
    molar_concentration: MOLAR_CONCENTRATION_CLASSES

    @staticmethod
    def create(raw_name: str, value: float) -> Analyte:
        if raw_name not in ANALYTE_MAPPINGS:
            msg = msg_for_error_on_unrecognized_value(
                "analyte name", raw_name, ANALYTE_MAPPINGS.keys()
            )
            raise AllotropeConversionError(msg)

        mapping = ANALYTE_MAPPINGS[raw_name]
        return Analyte(
            mapping["name"],
            MOLAR_CONCENTRATION_CLS_BY_UNIT[mapping["unit"]](value=value),
        )

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Analyte):
            return NotImplemented

        return self.name < other.name


@dataclass(frozen=True)
class Sample:
    identifier: str
    role_type: str
    measurement_time: str
    batch_identifier: Optional[str]
    analytes: list[Analyte]
    properties: dict[str, Any]

    @staticmethod
    def create(data: pd.Series[Any]) -> Sample:
        properties: dict[str, Any] = {
            property_name: property_dict["cls"](data[property_dict["col_name"]])
            for property_name, property_dict in PROPERTY_MAPPINGS.items()
            if property_dict["col_name"] in data
            and data[property_dict["col_name"]] is not None
        }

        batch_identifier = data.get("Batch ID")

        return Sample(
            identifier=data["Sample ID"],
            role_type=data["Sample Type"],
            measurement_time=data["Date & Time"].isoformat(),
            batch_identifier=str(batch_identifier) if batch_identifier else None,
            analytes=sorted(
                [
                    Analyte.create(raw_name, data[raw_name])
                    for raw_name in ANALYTE_MAPPINGS
                    if raw_name in data
                ]
            ),
            properties=properties,
        )


@dataclass(frozen=True)
class SampleList:
    analyst: str
    samples: list[Sample]

    @staticmethod
    def create(data: pd.DataFrame) -> SampleList:
        sample_data_rows = [row for _, row in data.iterrows()]

        if not sample_data_rows:
            msg = "Unable to find any sample."
            raise AllotropeConversionError(msg)

        analyst = sample_data_rows[0].get("Operator")

        if analyst is None:
            msg = "Unable to find the Operator."
            raise AllotropeConversionError(msg)

        return SampleList(
            str(analyst),
            [Sample.create(sample_data) for sample_data in sample_data_rows],
        )


@dataclass(frozen=True)
class Data:
    title: Title
    sample_list: SampleList

    @staticmethod
    def create(named_file_contents: NamedFileContents) -> Data:
        contents = named_file_contents.contents
        filename = named_file_contents.original_file_name
        data = read_csv(contents, parse_dates=["Date & Time"]).replace(np.nan, None)
        return Data(Title.create(filename), SampleList.create(data))



# header
# sample setup
# raw data
# amplification data
# multicomponent data
# results
# melt curve raw data

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from io import StringIO
import re
from typing import Optional

import numpy as np
import pandas as pd

from allotropy.allotrope.models.pcr_benchling_2023_09_qpcr import ExperimentType
from allotropy.allotrope.pandas_util import read_csv
from allotropy.parsers.appbio_quantstudio.calculated_document import CalculatedDocument
from allotropy.parsers.appbio_quantstudio.referenceable import Referenceable
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_empty_df,
    assert_not_none,
    df_to_series,
    try_bool_from_series_or_none,
    try_float_from_series,
    try_float_from_series_or_none,
    try_float_or_none,
    try_int,
    try_int_from_series,
    try_str_from_series,
    try_str_from_series_or_default,
    try_str_from_series_or_none,
)


@dataclass(frozen=True)
class Header:
    measurement_time: str
    plate_well_count: int
    experiment_type: ExperimentType
    device_identifier: str
    model_number: str
    device_serial_number: str
    measurement_method_identifier: str
    pcr_detection_chemistry: str
    passive_reference_dye_setting: Optional[str]
    barcode: Optional[str]
    analyst: Optional[str]
    experimental_data_identifier: Optional[str]

    @staticmethod
    def create(reader: LinesReader) -> Header:
        lines = [line.replace("*", "", 1) for line in reader.pop_until(r"^\[.+\]")]
        csv_stream = StringIO("\n".join(lines))
        raw_data = read_csv(csv_stream, header=None, sep="=", names=["index", "values"])
        data = pd.Series(raw_data["values"].values, index=raw_data["index"])
        data.index = data.index.str.strip()
        data = data.str.strip().replace("NA", None)

        experiments_type_options = {
            "Standard Curve": ExperimentType.standard_curve_qPCR_experiment,
            "Relative Standard Curve": ExperimentType.relative_standard_curve_qPCR_experiment,
            "Comparative Cт (ΔΔCт)": ExperimentType.comparative_CT_qPCR_experiment,
            "Melt Curve": ExperimentType.melt_curve_qPCR_experiment,
            "Genotyping": ExperimentType.genotyping_qPCR_experiment,
            "Presence/Absence": ExperimentType.presence_absence_qPCR_experiment,
        }

        return Header(
            measurement_time=try_str_from_series(data, "Experiment Run End Time"),
            plate_well_count=assert_not_none(
                try_int(
                    assert_not_none(
                        re.match(
                            "(96)|(384)",
                            try_str_from_series(data, "Block Type"),
                        ),
                        msg="Unable to find plate well count",
                    ).group(),
                    "plate well count",
                ),
                msg="Unable to interpret plate well count",
            ),
            experiment_type=assert_not_none(
                experiments_type_options.get(
                    try_str_from_series(data, "Experiment Type"),
                ),
                msg="Unable to find valid experiment type",
            ),
            device_identifier=(
                try_str_from_series_or_none(data, "Instrument Name") or "NA"
            ),
            model_number=try_str_from_series(data, "Instrument Type"),
            device_serial_number=try_str_from_series_or_none(
                data, "Instrument Serial Number"
            )
            or "NA",
            measurement_method_identifier=try_str_from_series(
                data, "Quantification Cycle Method"
            ),
            pcr_detection_chemistry=try_str_from_series(data, "Chemistry"),
            passive_reference_dye_setting=try_str_from_series_or_none(
                data, "Passive Reference"
            ),
            barcode=try_str_from_series_or_none(data, "Experiment Barcode"),
            analyst=try_str_from_series_or_none(data, "Experiment User Name"),
            experimental_data_identifier=try_str_from_series_or_none(
                data, "Experiment Name"
            ),
        )


@dataclass
class WellItem(Referenceable):
    identifier: int
    target_dna_description: str
    sample_identifier: str
    reporter_dye_setting: Optional[str]
    position: Optional[str]
    well_location_identifier: Optional[str]
    quencher_dye_setting: Optional[str]
    sample_role_type: Optional[str]
    _amplification_data: Optional[AmplificationData] = None
    _result: Optional[Result] = None

    @property
    def amplification_data(self) -> AmplificationData:
        return assert_not_none(
            self._amplification_data,
            msg=f"Unable to find amplification data for target '{self.target_dna_description}' in well {self.identifier} .",
        )

    @amplification_data.setter
    def amplification_data(self, amplification_data: AmplificationData) -> None:
        self._amplification_data = amplification_data

    @property
    def result(self) -> Result:
        return assert_not_none(
            self._result,
            msg=f"Unable to find result data for well {self.identifier}.",
        )

    @result.setter
    def result(self, result: Result) -> None:
        self._result = result

    @staticmethod
    def create_genotyping(data: pd.Series[str]) -> tuple[WellItem, WellItem]:
        identifier = try_int_from_series(data, "Well")

        snp_name = try_str_from_series(
            data,
            "SNP Assay Name",
            msg=f"Unable to find snp name for well {identifier}",
        )

        sample_identifier = try_str_from_series(
            data,
            "Sample Name",
            msg=f"Unable to find sample identifier for well {identifier}",
        )

        allele1 = try_str_from_series(
            data,
            "Allele1 Name",
            msg=f"Unable to find allele 1 for well {identifier}",
        )

        allele2 = try_str_from_series(
            data,
            "Allele2 Name",
            msg=f"Unable to find allele 2 for well {identifier}",
        )

        return (
            WellItem(
                uuid=random_uuid_str(),
                identifier=identifier,
                target_dna_description=f"{snp_name}-{allele1}",
                sample_identifier=sample_identifier,
                reporter_dye_setting=try_str_from_series_or_none(
                    data, "Allele1 Reporter"
                ),
                position=try_str_from_series_or_default(
                    data, "Well Position", default="UNDEFINED"
                ),
                well_location_identifier=try_str_from_series_or_none(
                    data, "Well Position"
                ),
                quencher_dye_setting=try_str_from_series_or_none(data, "Quencher"),
                sample_role_type=try_str_from_series_or_none(data, "Task"),
            ),
            WellItem(
                uuid=random_uuid_str(),
                identifier=identifier,
                target_dna_description=f"{snp_name}-{allele2}",
                sample_identifier=sample_identifier,
                reporter_dye_setting=try_str_from_series_or_none(
                    data, "Allele2 Reporter"
                ),
                position=try_str_from_series_or_default(
                    data, "Well Position", default="UNDEFINED"
                ),
                well_location_identifier=try_str_from_series_or_none(
                    data, "Well Position"
                ),
                quencher_dye_setting=try_str_from_series_or_none(data, "Quencher"),
                sample_role_type=try_str_from_series_or_none(data, "Task"),
            ),
        )

    @staticmethod
    def create_generic(data: pd.Series[str]) -> WellItem:
        identifier = try_int_from_series(data, "Well")

        target_dna_description = try_str_from_series(
            data,
            "Target Name",
            msg=f"Unable to find target dna description for well {identifier}",
        )

        sample_identifier = try_str_from_series(
            data,
            "Sample Name",
            msg=f"Unable to find sample identifier for well {identifier}",
        )

        return WellItem(
            uuid=random_uuid_str(),
            identifier=identifier,
            target_dna_description=target_dna_description,
            sample_identifier=sample_identifier,
            reporter_dye_setting=try_str_from_series_or_none(data, "Reporter"),
            position=try_str_from_series_or_default(
                data, "Well Position", default="UNDEFINED"
            ),
            well_location_identifier=try_str_from_series_or_none(data, "Well Position"),
            quencher_dye_setting=try_str_from_series_or_none(data, "Quencher"),
            sample_role_type=try_str_from_series_or_none(data, "Task"),
        )


@dataclass
class Well:
    identifier: int
    items: dict[str, WellItem]
    _multicomponent_data: Optional[MulticomponentData] = None
    _melt_curve_raw_data: Optional[MeltCurveRawData] = None
    _calculated_documents: Optional[list[CalculatedDocument]] = None

    def get_well_item(self, target: str) -> WellItem:
        well_item = self.items.get(target)
        return assert_not_none(
            well_item,
            msg=f"Unable to find target DNA '{target}' for well {self.identifier}.",
        )

    @property
    def multicomponent_data(self) -> Optional[MulticomponentData]:
        return self._multicomponent_data

    @multicomponent_data.setter
    def multicomponent_data(self, multicomponent_data: MulticomponentData) -> None:
        self._multicomponent_data = multicomponent_data

    @property
    def melt_curve_raw_data(self) -> Optional[MeltCurveRawData]:
        return self._melt_curve_raw_data

    @melt_curve_raw_data.setter
    def melt_curve_raw_data(self, melt_curve_raw_data: MeltCurveRawData) -> None:
        self._melt_curve_raw_data = melt_curve_raw_data

    @property
    def calculated_documents(self) -> list[CalculatedDocument]:
        return self._calculated_documents if self._calculated_documents else []

    def add_calculated_document(self, calculated_document: CalculatedDocument) -> None:
        if not self._calculated_documents:
            self._calculated_documents = []
        self._calculated_documents.append(calculated_document)

    @staticmethod
    def create_genotyping(identifier: int, well_data: pd.Series[str]) -> Well:
        return Well(
            identifier=identifier,
            items={
                well_item.target_dna_description: well_item
                for well_item in WellItem.create_genotyping(well_data)
            },
        )

    @staticmethod
    def create_generic(identifier: int, well_data: pd.DataFrame) -> Well:
        return Well(
            identifier=identifier,
            items={
                item_data["Target Name"]: WellItem.create_generic(item_data)
                for _, item_data in well_data.iterrows()
            },
        )


@dataclass(frozen=True)
class WellList:
    wells: list[Well]

    def iter_well_items(self) -> Iterator[WellItem]:
        for well in self.wells:
            yield from well.items.values()

    def __iter__(self) -> Iterator[Well]:
        return iter(self.wells)

    @staticmethod
    def create(reader: LinesReader, experiment_type: ExperimentType) -> WellList:
        assert_not_none(
            reader.drop_until(r"^\[Sample Setup\]"),
            msg="Unable to find 'Sample Setup' section in file.",
        )

        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        raw_data = read_csv(csv_stream, sep="\t").replace(np.nan, None)
        data = raw_data[raw_data["Sample Name"].notnull()]

        if experiment_type == ExperimentType.genotyping_qPCR_experiment:
            return WellList(
                [
                    Well.create_genotyping(
                        try_int(str(identifier), "genotyping well identifier"),
                        well_data,
                    )
                    for identifier, well_data in data.iterrows()
                ]
            )
        return WellList(
            [
                Well.create_generic(
                    try_int(str(identifier), "well identifier"),
                    well_data,
                )
                for identifier, well_data in data.groupby("Well")
            ]
        )


@dataclass(frozen=True)
class RawData:
    lines: list[str]

    @staticmethod
    def create(reader: LinesReader) -> Optional[RawData]:
        if reader.match(r"^\[Raw Data\]"):
            reader.pop()  # remove title
            return RawData(lines=list(reader.pop_until(r"^\[.+\]")))
        return None


@dataclass(frozen=True)
class AmplificationData:
    total_cycle_number_setting: float
    cycle: list[float]
    rn: list[Optional[float]]
    delta_rn: list[Optional[float]]

    @staticmethod
    def get_data(reader: LinesReader) -> pd.DataFrame:
        assert_not_none(
            reader.drop_until(r"^\[Amplification Data\]"),
            msg="Unable to find 'Amplification Data' section in file.",
        )

        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        return read_csv(csv_stream, sep="\t", thousands=r",")

    @staticmethod
    def create(
        amplification_data: pd.DataFrame, well_item: WellItem
    ) -> AmplificationData:
        well_data = assert_not_empty_df(
            amplification_data[amplification_data["Well"] == well_item.identifier],
            msg=f"Unable to find amplification data for well {well_item.identifier}.",
        )

        target_data = assert_not_empty_df(
            well_data[well_data["Target Name"] == well_item.target_dna_description],
            msg=f"Unable to find amplification data for target '{well_item.target_dna_description}' in well {well_item.identifier} .",
        )

        return AmplificationData(
            total_cycle_number_setting=float(target_data["Cycle"].max()),
            cycle=target_data["Cycle"].tolist(),
            rn=target_data["Rn"].tolist(),
            delta_rn=target_data["Delta Rn"].tolist(),
        )


@dataclass(frozen=True)
class MulticomponentData:
    cycle: list[float]
    columns: dict[str, list[Optional[float]]]

    def get_column(self, name: str) -> list[Optional[float]]:
        return assert_not_none(
            self.columns.get(name),
            msg=f"Unable to obtain '{name}' from multicomponent data.",
        )

    @staticmethod
    def get_data(reader: LinesReader) -> Optional[pd.DataFrame]:
        if not reader.match(r"^\[Multicomponent Data\]"):
            return None
        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        return read_csv(csv_stream, sep="\t", thousands=r",")

    @staticmethod
    def create(data: pd.DataFrame, well: Well) -> MulticomponentData:
        well_data = assert_not_empty_df(
            data[data["Well"] == well.identifier],
            msg=f"Unable to find multi component data for well {well.identifier}.",
        )

        return MulticomponentData(
            cycle=well_data["Cycle"].tolist(),
            columns={
                name: well_data[name].tolist()  # type: ignore[misc]
                for name in well_data
                if name not in ["Well", "Cycle", "Well Position"]
            },
        )


@dataclass(frozen=True)
class Result:
    cycle_threshold_value_setting: float
    cycle_threshold_result: Optional[float]
    automatic_cycle_threshold_enabled_setting: Optional[bool]
    automatic_baseline_determination_enabled_setting: Optional[bool]
    normalized_reporter_result: Optional[float]
    baseline_corrected_reporter_result: Optional[float]
    genotyping_determination_result: Optional[str]
    genotyping_determination_method_setting: Optional[float]
    quantity: Optional[float]
    quantity_mean: Optional[float]
    quantity_sd: Optional[float]
    ct_mean: Optional[float]
    ct_sd: Optional[float]
    delta_ct_mean: Optional[float]
    delta_ct_se: Optional[float]
    delta_delta_ct: Optional[float]
    rq: Optional[float]
    rq_min: Optional[float]
    rq_max: Optional[float]
    rn_mean: Optional[float]
    rn_sd: Optional[float]
    y_intercept: Optional[float]
    r_squared: Optional[float]
    slope: Optional[float]
    efficiency: Optional[float]

    @staticmethod
    def get_data(reader: LinesReader) -> tuple[pd.DataFrame, pd.Series[str]]:
        assert_not_none(
            reader.drop_until(r"^\[Results\]"),
            msg="Unable to find 'Results' section in file.",
        )

        reader.pop()  # remove title
        data_lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(data_lines))
        data = read_csv(csv_stream, sep="\t", thousands=r",").replace(np.nan, None)

        reader.drop_empty()

        if reader.match(r"\[.+\]"):
            return data, pd.Series()

        metadata_lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(metadata_lines))
        raw_data = read_csv(csv_stream, header=None, sep="=", names=["index", "values"])
        metadata = pd.Series(raw_data["values"].values, index=raw_data["index"])
        metadata.index = metadata.index.str.strip()

        reader.drop_empty()

        return data, metadata.str.strip()

    @staticmethod
    def create_genotyping(data: pd.DataFrame, well_item: WellItem) -> Result:
        well_data = assert_not_empty_df(
            data[data["Well"] == well_item.identifier],
            msg=f"Unable to find result data for well {well_item.identifier}.",
        )

        snp_assay_name, _ = well_item.target_dna_description.split("-")
        target_data = df_to_series(
            assert_not_empty_df(
                well_data[well_data["SNP Assay Name"] == snp_assay_name],
                msg=f"Unable to find result data for well {well_item.identifier}.",
            ),
            msg=f"Expected exactly 1 row of results to be associated with target '{well_item.target_dna_description}' in well {well_item.identifier}.",
        )

        _, raw_allele = well_item.target_dna_description.split("-")
        allele = raw_allele.replace(" ", "")
        cycle_threshold_value_setting = try_float_from_series(
            target_data,
            f"{allele} Ct Threshold",
            msg=f"Unable to find cycle threshold value setting for well {well_item.identifier}",
        )

        cycle_threshold_result = assert_not_none(
            target_data.get(f"{allele} Ct"),
            msg="Unable to find cycle threshold result",
        )

        return Result(
            cycle_threshold_value_setting=cycle_threshold_value_setting,
            cycle_threshold_result=try_float_or_none(str(cycle_threshold_result)),
            automatic_cycle_threshold_enabled_setting=try_bool_from_series_or_none(
                target_data, f"{allele} Automatic Ct Threshold"
            ),
            automatic_baseline_determination_enabled_setting=try_bool_from_series_or_none(
                target_data, f"{allele} Automatic Baseline"
            ),
            normalized_reporter_result=try_float_from_series_or_none(target_data, "Rn"),
            baseline_corrected_reporter_result=try_float_from_series_or_none(
                target_data, f"{allele} Delta Rn"
            ),
            genotyping_determination_result=try_str_from_series_or_none(
                target_data, "Call"
            ),
            genotyping_determination_method_setting=try_float_from_series_or_none(
                target_data, "Threshold Value"
            ),
            quantity=try_float_from_series_or_none(target_data, "Quantity"),
            quantity_mean=try_float_from_series_or_none(target_data, "Quantity Mean"),
            quantity_sd=try_float_from_series_or_none(target_data, "Quantity SD"),
            ct_mean=try_float_from_series_or_none(target_data, "Ct Mean"),
            ct_sd=try_float_from_series_or_none(target_data, "Ct SD"),
            delta_ct_mean=try_float_from_series_or_none(target_data, "Delta Ct Mean"),
            delta_ct_se=try_float_from_series_or_none(target_data, "Delta Ct SE"),
            delta_delta_ct=try_float_from_series_or_none(target_data, "Delta Delta Ct"),
            rq=try_float_from_series_or_none(target_data, "RQ"),
            rq_min=try_float_from_series_or_none(target_data, "RQ Min"),
            rq_max=try_float_from_series_or_none(target_data, "RQ Max"),
            rn_mean=try_float_from_series_or_none(target_data, "Rn Mean"),
            rn_sd=try_float_from_series_or_none(target_data, "Rn SD"),
            y_intercept=try_float_from_series_or_none(target_data, "Y-Intercept"),
            r_squared=try_float_from_series_or_none(target_data, "R(superscript 2)"),
            slope=try_float_from_series_or_none(target_data, "Slope"),
            efficiency=try_float_from_series_or_none(target_data, "Efficiency"),
        )

    @staticmethod
    def create_generic(data: pd.DataFrame, well_item: WellItem) -> Result:
        well_data = assert_not_empty_df(
            data[data["Well"] == well_item.identifier],
            msg=f"Unable to find result data for well {well_item.identifier}.",
        )

        target_data = df_to_series(
            assert_not_empty_df(
                well_data[well_data["Target Name"] == well_item.target_dna_description],
                msg=f"Unable to find result data for well {well_item.identifier}.",
            ),
            msg=f"Expected exactly 1 row of results to be associated with target '{well_item.target_dna_description}' in well {well_item.identifier}.",
        )

        cycle_threshold_result = assert_not_none(
            target_data.get("CT"),
            msg="Unable to find cycle threshold result",
        )

        return Result(
            cycle_threshold_value_setting=try_float_from_series(
                target_data,
                "Ct Threshold",
                msg=f"Unable to find cycle threshold value setting for well {well_item.identifier}",
            ),
            cycle_threshold_result=try_float_or_none(str(cycle_threshold_result)),
            automatic_cycle_threshold_enabled_setting=try_bool_from_series_or_none(
                target_data, "Automatic Ct Threshold"
            ),
            automatic_baseline_determination_enabled_setting=try_bool_from_series_or_none(
                target_data, "Automatic Baseline"
            ),
            normalized_reporter_result=try_float_from_series_or_none(target_data, "Rn"),
            baseline_corrected_reporter_result=try_float_from_series_or_none(
                target_data, "Delta Rn"
            ),
            genotyping_determination_result=try_str_from_series_or_none(
                target_data, "Call"
            ),
            genotyping_determination_method_setting=try_float_from_series_or_none(
                target_data, "Threshold Value"
            ),
            quantity=try_float_from_series_or_none(target_data, "Quantity"),
            quantity_mean=try_float_from_series_or_none(target_data, "Quantity Mean"),
            quantity_sd=try_float_from_series_or_none(target_data, "Quantity SD"),
            ct_mean=try_float_from_series_or_none(target_data, "Ct Mean"),
            ct_sd=try_float_from_series_or_none(target_data, "Ct SD"),
            delta_ct_mean=try_float_from_series_or_none(target_data, "Delta Ct Mean"),
            delta_ct_se=try_float_from_series_or_none(target_data, "Delta Ct SE"),
            delta_delta_ct=try_float_from_series_or_none(target_data, "Delta Delta Ct"),
            rq=try_float_from_series_or_none(target_data, "RQ"),
            rq_min=try_float_from_series_or_none(target_data, "RQ Min"),
            rq_max=try_float_from_series_or_none(target_data, "RQ Max"),
            rn_mean=try_float_from_series_or_none(target_data, "Rn Mean"),
            rn_sd=try_float_from_series_or_none(target_data, "Rn SD"),
            y_intercept=try_float_from_series_or_none(target_data, "Y-Intercept"),
            r_squared=try_float_from_series_or_none(target_data, "R(superscript 2)"),
            slope=try_float_from_series_or_none(target_data, "Slope"),
            efficiency=try_float_from_series_or_none(target_data, "Efficiency"),
        )

    @staticmethod
    def create(
        data: pd.DataFrame, well_item: WellItem, experiment_type: ExperimentType
    ) -> Result:
        if experiment_type == ExperimentType.genotyping_qPCR_experiment:
            return Result.create_genotyping(data, well_item)
        return Result.create_generic(data, well_item)


@dataclass(frozen=True)
class MeltCurveRawData:
    reading: list[float]
    fluorescence: list[Optional[float]]
    derivative: list[Optional[float]]

    @staticmethod
    def create(data: pd.DataFrame, well: Well) -> MeltCurveRawData:
        well_data = assert_not_empty_df(
            data[data["Well"] == well.identifier],
            msg=f"Unable to find melt curve raw data for well {well.identifier}.",
        )
        return MeltCurveRawData(
            reading=well_data["Reading"].tolist(),
            fluorescence=well_data["Fluorescence"].tolist(),
            derivative=well_data["Derivative"].tolist(),
        )

    @staticmethod
    def get_data(reader: LinesReader) -> Optional[pd.DataFrame]:
        if not reader.match(r"^\[Melt Curve Raw Data\]"):
            return None
        reader.pop()  # remove title
        lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(lines))
        return read_csv(csv_stream, sep="\t", thousands=r",")


@dataclass(frozen=True)
class Data:
    header: Header
    wells: WellList
    raw_data: Optional[RawData]
    endogenous_control: str
    reference_sample: str
    calculated_documents: list[CalculatedDocument]



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.lines_reader import CsvReader

EMPTY_CSV_LINE = r"^,*$"
PROTOCOL_ID = "Weyland Yutani Example"
ASSAY_ID = "Example Assay"


@dataclass(frozen=True)
class BasicAssayInfo:
    protocol_id: str
    assay_id: str
    checksum: Optional[str]

    @staticmethod
    def create(bottom: Optional[pd.DataFrame]) -> BasicAssayInfo:
        checksum = (
            None
            if (bottom is None) or (bottom.iloc[0, 0] != "Checksum")
            else str(bottom.iloc[0, 1])
        )
        return BasicAssayInfo(
            protocol_id=PROTOCOL_ID,
            assay_id=ASSAY_ID,
            checksum=checksum,
        )


@dataclass(frozen=True)
class Instrument:
    serial_number: str
    nickname: str

    # TODO: extract and fill in real values for serial number and nickname
    @staticmethod
    def create() -> Instrument:
        return Instrument(serial_number="", nickname="")


@dataclass(frozen=True)
class Result:
    col: str
    row: str
    value: float


@dataclass(frozen=True)
class Plate:
    number: str
    results: list[Result]

    @staticmethod
    def create(df: Optional[pd.DataFrame]) -> list[Plate]:
        if df is None:
            return []
        pivoted = df.T
        if pivoted.iloc[1, 0] != "A":
            msg = "Column header(s) not found."
            raise AllotropeConversionError(msg)
        stripped = pivoted.drop(0, axis=0).drop(0, axis=1)
        rows, cols = stripped.shape
        stripped.index = [df.iloc[0, i + 1] for i in range(rows)]  # type: ignore
        stripped.columns = [str(int(df.iloc[i, 0])) for i in range(1, cols + 1)]  # type: ignore
        return [
            Plate(
                number="0",
                results=[
                    Result(col, row, float(stripped.loc[col, row]))
                    for col, row in stripped.stack().index
                ],
            )
        ]


@dataclass(frozen=True)
class Data:
    plates: list[Plate]
    number_of_wells: Optional[float]
    basic_assay_info: BasicAssayInfo
    instrument: Instrument

    @staticmethod
    def create(reader: CsvReader) -> Data:
        _ = reader.pop_csv_block_as_df(empty_pat=EMPTY_CSV_LINE)
        middle = reader.pop_csv_block_as_df(empty_pat=EMPTY_CSV_LINE)
        bottom = reader.pop_csv_block_as_df(empty_pat=EMPTY_CSV_LINE)

        plates = Plate.create(middle)
        return Data(
            basic_assay_info=BasicAssayInfo.create(bottom),
            plates=plates,
            number_of_wells=len(plates[0].results),
            instrument=Instrument.create(),
        )



from collections.abc import Sequence
from itertools import chain
import math
from typing import Optional, Union

from allotropy.allotrope.models.plate_reader_benchling_2023_09_plate_reader import (
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    ContainerType,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceSystemDocument,
    FluorescencePointDetectionDeviceControlAggregateDocument,
    FluorescencePointDetectionDeviceControlDocumentItem,
    FluorescencePointDetectionMeasurementDocumentItems,
    LuminescencePointDetectionDeviceControlAggregateDocument,
    LuminescencePointDetectionDeviceControlDocumentItem,
    LuminescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    OpticalImagingMeasurementDocumentItems,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    SampleDocument,
    ScanPositionSettingPlateReader,
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument,
    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueRelativeLightUnit,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
    JsonFloat,
    TQuantityValue,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.exceptions import (
    AllotropeConversionError,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.moldev_softmax_pro.constants import (
    DEVICE_TYPE,
    EPOCH,
    NULL,
    REDUCED,
)
from allotropy.parsers.moldev_softmax_pro.softmax_pro_structure import (
    Data,
    GroupBlock,
    GroupSampleData,
    PlateBlock,
    ScanPosition,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
)
from allotropy.parsers.vendor_parser import VendorParser

MeasurementDocumentItems = Union[
    OpticalImagingMeasurementDocumentItems,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
    FluorescencePointDetectionMeasurementDocumentItems,
    LuminescencePointDetectionMeasurementDocumentItems,
]


def to_json_float(value: float) -> JsonFloat:
    return InvalidJsonFloat.NaN if math.isnan(value) else value


class SoftmaxproParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        lines = read_to_lines(named_file_contents)
        reader = CsvReader(lines)
        data = Data.create(reader)
        return self._get_model(named_file_contents.original_file_name, data)

    def _get_model(self, file_name: str, data: Data) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=NULL,
                    model_number=NULL,
                ),
                data_system_document=DataSystemDocument(
                    file_name=file_name,
                    software_name="SoftMax Pro",
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                plate_reader_document=[
                    self._get_plate_reader_document_item(plate_block, position)
                    for plate_block in data.block_list.plate_blocks.values()
                    for position in plate_block.iter_wells()
                ],
                calculated_data_aggregate_document=self._get_calc_docs(data),
            ),
        )

    def _get_plate_reader_document_item(
        self, plate_block: PlateBlock, position: str
    ) -> PlateReaderDocumentItem:
        plate_block_type = plate_block.get_plate_block_type()

        measurement_document: Sequence[MeasurementDocumentItems]

        if plate_block_type == "Absorbance":
            measurement_document = self._get_absorbance_measurement_document(
                plate_block,
                position,
            )
        elif plate_block_type == "Luminescence":
            measurement_document = self._get_luminescence_measurement_document(
                plate_block,
                position,
            )
        elif plate_block_type == "Fluorescence":
            measurement_document = self._get_fluorescence_measurement_document(
                plate_block,
                position,
            )
        else:
            error = f"{plate_block_type} is not a valid plate block type."
            raise AllotropeConversionError(error)

        return PlateReaderDocumentItem(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_time=EPOCH,
                plate_well_count=TQuantityValueNumber(plate_block.header.num_wells),
                container_type=ContainerType.well_plate,
                measurement_document=list(measurement_document),
            )
        )

    def _get_fluorescence_measurement_document(
        self, plate_block: PlateBlock, position: str
    ) -> list[FluorescencePointDetectionMeasurementDocumentItems]:
        return [
            FluorescencePointDetectionMeasurementDocumentItems(
                measurement_identifier=data_element.uuid,
                fluorescence=TQuantityValueRelativeFluorescenceUnit(
                    value=data_element.value
                ),
                compartment_temperature=(
                    None
                    if data_element.temperature is None
                    else TQuantityValueDegreeCelsius(
                        to_json_float(data_element.temperature)
                    )
                ),
                sample_document=SampleDocument(
                    location_identifier=data_element.position,
                    well_plate_identifier=plate_block.header.name,
                    sample_identifier=data_element.sample_identifier,
                ),
                device_control_aggregate_document=FluorescencePointDetectionDeviceControlAggregateDocument(
                    device_control_document=[
                        FluorescencePointDetectionDeviceControlDocumentItem(
                            device_type=DEVICE_TYPE,
                            detection_type=plate_block.header.read_mode,
                            scan_position_setting__plate_reader_=(
                                ScanPositionSettingPlateReader.top_scan_position__plate_reader_
                                if plate_block.header.scan_position == ScanPosition.TOP
                                else ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_
                            ),
                            detector_wavelength_setting=TQuantityValueNanometer(
                                data_element.wavelength
                            ),
                            excitation_wavelength_setting=(
                                None
                                if plate_block.header.excitation_wavelengths is None
                                else TQuantityValueNanometer(
                                    plate_block.header.excitation_wavelengths[idx]
                                )
                            ),
                            wavelength_filter_cutoff_setting=(
                                None
                                if plate_block.header.cutoff_filters is None
                                else TQuantityValueNanometer(
                                    plate_block.header.cutoff_filters[idx]
                                )
                            ),
                            number_of_averages=TQuantityValueNumber(
                                plate_block.header.reads_per_well,
                            ),
                            detector_gain_setting=plate_block.header.pmt_gain,
                        )
                    ]
                ),
            )
            for idx, data_element in enumerate(plate_block.iter_data_elements(position))
        ]

    def _get_luminescence_measurement_document(
        self, plate_block: PlateBlock, position: str
    ) -> list[LuminescencePointDetectionMeasurementDocumentItems]:
        reads_per_well = assert_not_none(
            plate_block.header.reads_per_well,
            msg="Unable to find plate block reads per well.",
        )

        return [
            LuminescencePointDetectionMeasurementDocumentItems(
                measurement_identifier=data_element.uuid,
                luminescence=TQuantityValueRelativeLightUnit(value=data_element.value),
                compartment_temperature=(
                    None
                    if data_element.temperature is None
                    else TQuantityValueDegreeCelsius(
                        to_json_float(data_element.temperature)
                    )
                ),
                sample_document=SampleDocument(
                    location_identifier=data_element.position,
                    well_plate_identifier=plate_block.header.name,
                    sample_identifier=data_element.sample_identifier,
                ),
                device_control_aggregate_document=LuminescencePointDetectionDeviceControlAggregateDocument(
                    device_control_document=[
                        LuminescencePointDetectionDeviceControlDocumentItem(
                            device_type=DEVICE_TYPE,
                            detection_type=plate_block.header.read_mode,
                            detector_wavelength_setting=TQuantityValueNanometer(
                                data_element.wavelength
                            ),
                            number_of_averages=TQuantityValueNumber(reads_per_well),
                            detector_gain_setting=plate_block.header.pmt_gain,
                        )
                    ]
                ),
            )
            for data_element in plate_block.iter_data_elements(position)
        ]

    def _get_absorbance_measurement_document(
        self, plate_block: PlateBlock, position: str
    ) -> list[UltravioletAbsorbancePointDetectionMeasurementDocumentItems]:
        return [
            UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
                measurement_identifier=data_element.uuid,
                absorbance=TQuantityValueMilliAbsorbanceUnit(value=data_element.value),
                compartment_temperature=(
                    None
                    if data_element.temperature is None
                    else TQuantityValueDegreeCelsius(
                        to_json_float(data_element.temperature)
                    )
                ),
                sample_document=SampleDocument(
                    location_identifier=data_element.position,
                    well_plate_identifier=plate_block.header.name,
                    sample_identifier=data_element.sample_identifier,
                ),
                device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                    device_control_document=[
                        UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
                            device_type=DEVICE_TYPE,
                            detection_type=plate_block.header.read_mode,
                            detector_wavelength_setting=TQuantityValueNanometer(
                                data_element.wavelength
                            ),
                        )
                    ]
                ),
            )
            for data_element in plate_block.iter_data_elements(position)
        ]

    def _get_calc_docs(self, data: Data) -> Optional[CalculatedDataAggregateDocument]:
        calc_docs = self._get_reduced_calc_docs(data) + self._get_group_calc_docs(data)
        return CalculatedDataAggregateDocument(calc_docs) if calc_docs else None

    def _get_calc_docs_data_sources(
        self, plate_block: PlateBlock, position: str
    ) -> list[DataSourceDocumentItem]:
        return [
            DataSourceDocumentItem(
                data_source_identifier=data_source.uuid,
                data_source_feature=plate_block.get_plate_block_type(),
            )
            for data_source in plate_block.iter_data_elements(position)
        ]

    def _build_calc_doc(
        self,
        name: str,
        value: float,
        data_sources: list[DataSourceDocumentItem],
        description: Optional[str] = None,
    ) -> CalculatedDataDocumentItem:
        return CalculatedDataDocumentItem(
            calculated_data_identifier=random_uuid_str(),
            calculated_data_name=name,
            calculation_description=description,
            calculated_result=TQuantityValue(
                unit=UNITLESS,
                value=value,
            ),
            data_source_aggregate_document=DataSourceAggregateDocument(
                data_source_document=data_sources,
            ),
        )

    def _get_reduced_calc_docs(self, data: Data) -> list[CalculatedDataDocumentItem]:
        return [
            self._build_calc_doc(
                name=REDUCED,
                value=reduced_data_element.value,
                data_sources=self._get_calc_docs_data_sources(
                    plate_block,
                    reduced_data_element.position,
                ),
            )
            for plate_block in data.block_list.plate_blocks.values()
            for reduced_data_element in plate_block.iter_reduced_data()
        ]

    def _get_group_agg_calc_docs(
        self, data: Data, group_block: GroupBlock, group_sample_data: GroupSampleData
    ) -> list[CalculatedDataDocumentItem]:
        return [
            self._build_calc_doc(
                name=aggregated_entry.name,
                value=aggregated_entry.value,
                data_sources=list(
                    chain.from_iterable(
                        self._get_calc_docs_data_sources(
                            data.block_list.plate_blocks[group_data_element.plate],
                            group_data_element.position,
                        )
                        for group_data_element in group_sample_data.data_elements
                    )
                ),
                description=group_block.group_columns.data.get(aggregated_entry.name),
            )
            for aggregated_entry in group_sample_data.aggregated_entries
        ]

    def _get_group_simple_calc_docs(
        self, data: Data, group_block: GroupBlock, group_sample_data: GroupSampleData
    ) -> list[CalculatedDataDocumentItem]:
        calculated_documents = []
        for group_data_element in group_sample_data.data_elements:
            data_sources = self._get_calc_docs_data_sources(
                data.block_list.plate_blocks[group_data_element.plate],
                group_data_element.position,
            )
            for entry in group_data_element.entries:
                calculated_documents.append(
                    self._build_calc_doc(
                        name=entry.name,
                        value=entry.value,
                        data_sources=data_sources,
                        description=group_block.group_columns.data.get(entry.name),
                    )
                )
        return calculated_documents

    def _get_group_calc_docs(self, data: Data) -> list[CalculatedDataDocumentItem]:
        calculated_documents = []
        for group_block in data.block_list.group_blocks:
            for group_sample_data in group_block.group_data.sample_data:
                calculated_documents += self._get_group_agg_calc_docs(
                    data, group_block, group_sample_data
                )
                calculated_documents += self._get_group_simple_calc_docs(
                    data, group_block, group_sample_data
                )
        return calculated_documents



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional

import numpy as np
import pandas as pd

from allotropy.allotrope.models.cell_counting_benchling_2023_11_cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDetectorDeviceControlAggregateDocument,
    CellCountingDetectorMeasurementDocumentItem,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlDocumentItemModel,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    Model,
    ProcessedDataAggregateDocument1,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_vi_cell_blu.constants import (
    DEFAULT_ANALYST,
    DEFAULT_MODEL_NUMBER,
    VICELL_BLU_SOFTWARE_NAME,
)
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_reader import ViCellBluReader
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser


class SampleProperty(Enum):
    AVERAGE_VIABLE_DIAMETER = ("Average viable diameter (μm)", TQuantityValueMicrometer)
    AVERAGE_CIRCULARITY = ("Average circularity", TQuantityValueUnitless)
    AVERAGE_DIAMETER = ("Average diameter (μm)", TQuantityValueMicrometer)
    AVERAGE_VIABLE_CIRCULARITY = ("Average viable circularity", TQuantityValueUnitless)
    DILUTION = ("Dilution", TQuantityValueUnitless)
    MAXIMUM_DIAMETER = ("Maximum Diameter (μm)", TQuantityValueMicrometer)
    MINIMUM_DIAMETER = ("Minimum Diameter (μm)", TQuantityValueMicrometer)
    CELL_COUNT = ("Cell count", TQuantityValueCell)
    TOTAL_CELLS_ML = ("Total (x10^6) cells/mL", TQuantityValueMillionCellsPerMilliliter)
    VIABILITY = ("Viability (%)", TQuantityValuePercent)
    VIABLE_CELLS = ("Viable cells", TQuantityValueCell)
    VIABLE_CELLS_ML = (
        "Viable (x10^6) cells/mL",
        TQuantityValueMillionCellsPerMilliliter,
    )

    def __init__(self, column_name: str, data_type: Any) -> None:
        self.column_name: str = column_name
        self.data_type: Any = data_type


@dataclass(frozen=True)
class _Sample:
    data_frame: pd.DataFrame
    row: int

    def get_value(self, column: str) -> Optional[Any]:
        if column not in self.data_frame.columns:
            return None
        value = self.data_frame[column][self.row]

        # https://stackoverflow.com/questions/50916422/python-typeerror-object-of-type-int64-is-not-json-serializable
        if isinstance(value, np.int64):
            return int(value)

        return value

    def get_value_not_none(self, column: str) -> Any:
        value = self.get_value(column)
        if value is None:
            msg = f"Missing value for column '{column}'."
            raise AllotropeConversionError(msg)
        return value

    def get_property_value(self, sample_property: SampleProperty) -> Any:
        return (
            sample_property.data_type(value=value)
            if (value := self.get_value(sample_property.column_name))
            else None
        )


class ViCellBluParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        return self._get_model(
            data=ViCellBluReader.read(named_file_contents),
            filename=named_file_contents.original_file_name,
        )

    def _get_model(self, data: pd.DataFrame, filename: str) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/11/cell-counting.manifest",
            cell_counting_aggregate_document=CellCountingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=DEFAULT_MODEL_NUMBER,
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name=VICELL_BLU_SOFTWARE_NAME,
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                cell_counting_document=self._get_cell_counting_document(data),
            ),
        )

    def _get_cell_counting_document(
        self, data: pd.DataFrame
    ) -> list[CellCountingDocumentItem]:
        return [
            self._get_cell_counting_document_item(_Sample(data, i))
            for i in range(len(data.index))
            if (_Sample(data, i).get_value("Cell count"))
        ]

    def _get_cell_counting_document_item(
        self, sample: _Sample
    ) -> CellCountingDocumentItem:
        return CellCountingDocumentItem(
            analyst=sample.get_value("Analysis by") or DEFAULT_ANALYST,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    CellCountingDetectorMeasurementDocumentItem(
                        measurement_time=self._get_date_time(
                            sample.get_value_not_none("Analysis date/time")
                        ),
                        measurement_identifier=random_uuid_str(),
                        sample_document=SampleDocument(sample_identifier=sample.get_value("Sample ID")),  # type: ignore[arg-type]
                        device_control_aggregate_document=CellCountingDetectorDeviceControlAggregateDocument(
                            device_control_document=[
                                DeviceControlDocumentItemModel(
                                    device_type="brightfield imager (cell counter)",
                                    detection_type="brightfield",
                                )
                            ]
                        ),
                        processed_data_aggregate_document=ProcessedDataAggregateDocument1(
                            processed_data_document=[
                                ProcessedDataDocumentItem(
                                    data_processing_document=DataProcessingDocument(
                                        cell_type_processing_method=sample.get_value(
                                            "Cell type"
                                        ),
                                        minimum_cell_diameter_setting=sample.get_property_value(
                                            SampleProperty.MINIMUM_DIAMETER
                                        ),
                                        maximum_cell_diameter_setting=sample.get_property_value(
                                            SampleProperty.MAXIMUM_DIAMETER
                                        ),
                                        cell_density_dilution_factor=sample.get_property_value(
                                            SampleProperty.DILUTION
                                        ),
                                    ),
                                    viability__cell_counter_=sample.get_property_value(
                                        SampleProperty.VIABILITY
                                    ),
                                    viable_cell_density__cell_counter_=sample.get_property_value(
                                        SampleProperty.VIABLE_CELLS_ML
                                    ),
                                    total_cell_count=sample.get_property_value(
                                        SampleProperty.CELL_COUNT
                                    ),
                                    total_cell_density__cell_counter_=sample.get_property_value(
                                        SampleProperty.TOTAL_CELLS_ML
                                    ),
                                    average_total_cell_diameter=sample.get_property_value(
                                        SampleProperty.AVERAGE_DIAMETER
                                    ),
                                    average_live_cell_diameter__cell_counter_=sample.get_property_value(
                                        SampleProperty.AVERAGE_VIABLE_DIAMETER
                                    ),
                                    viable_cell_count=sample.get_property_value(
                                        SampleProperty.VIABLE_CELLS
                                    ),
                                    average_total_cell_circularity=sample.get_property_value(
                                        SampleProperty.AVERAGE_CIRCULARITY
                                    ),
                                    average_viable_cell_circularity=sample.get_property_value(
                                        SampleProperty.AVERAGE_VIABLE_CIRCULARITY
                                    ),
                                ),
                            ]
                        ),
                    )
                ],
            ),
        )



#!/usr/bin/env python

"""Show CSV file as JSON."""

import argparse
import json

from allotropy.parser_factory import Vendor
from allotropy.to_allotrope import allotrope_from_io


def main() -> None:
    """Main driver."""
    args = parse_args()
    vendor = Vendor[args.vendor]
    with open(args.infile) as reader:
        asm_schema = allotrope_from_io(reader, args.infile, vendor)
    if args.outfile:
        with open(args.outfile, "w") as out:
            json.dump(asm_schema, out, indent=4)
    print(json.dumps(asm_schema, indent=4))  # noqa: T201


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--infile", type=str, required=True, help="file to convert")
    parser.add_argument("--vendor", type=str, required=True, help="vendor identifier")
    parser.add_argument(
        "--outfile", type=str, required=False, help="file to write results to"
    )
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



# generated by datamodel-codegen:
#   filename:  dpcr.json
#   timestamp: 2024-01-12T14:18:33+00:00

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMicroliter,
    TQuantityValueNumber,
    TQuantityValueNumberPerMicroliter,
    TQuantityValueSecondTime,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDateTimeValue,
    TQuantityValue,
    TStringValue,
)


class ExperimentType(Enum):
    dPCR_experiment = "dPCR experiment"


class ContainerType(Enum):
    reactor = "reactor"
    controlled_lab_reactor = "controlled lab reactor"
    tube = "tube"
    well_plate = "well plate"
    differential_scanning_calorimetry_pan = "differential scanning calorimetry pan"
    qPCR_reaction_block = "qPCR reaction block"
    vial_rack = "vial rack"
    pan = "pan"
    reservoir = "reservoir"
    array_card_block = "array card block"
    capillary = "capillary"
    disintegration_apparatus_basket = "disintegration apparatus basket"
    jar = "jar"
    container = "container"
    tray = "tray"
    basket = "basket"
    cell_holder = "cell holder"


@dataclass
class DataSourceDocumentItem:
    data_source_identifier: Optional[TStringValue] = None
    data_source_feature: Optional[TStringValue] = None


@dataclass
class DataSourceAggregateDocument:
    data_source_document: Optional[list[DataSourceDocumentItem]] = None


@dataclass
class DeviceSystemDocument:
    device_identifier: TStringValue
    asset_management_identifier: Optional[TStringValue] = None
    model_number: Optional[TStringValue] = None
    device_serial_number: Optional[TStringValue] = None
    firmware_version: Optional[TStringValue] = None
    description: Optional[TStringValue] = None
    brand_name: Optional[TStringValue] = None
    product_manufacturer: Optional[TStringValue] = None


@dataclass
class DataSystemDocument:
    data_system_instance_identifier: Optional[TStringValue] = None
    file_name: Optional[TStringValue] = None
    UNC_path: Optional[TStringValue] = None
    software_name: Optional[TStringValue] = None
    software_version: Optional[TStringValue] = None
    ASM_converter_name: Optional[TStringValue] = None
    ASM_converter_version: Optional[TStringValue] = None


@dataclass
class SampleDocument:
    sample_identifier: TStringValue
    batch_identifier: Optional[TStringValue] = None
    sample_role_type: Optional[TStringValue] = None
    well_location_identifier: Optional[TStringValue] = None
    well_plate_identifier: Optional[TStringValue] = None


@dataclass
class DeviceControlDocumentItem:
    device_type: TStringValue
    device_identifier: Optional[TStringValue] = None
    detection_type: Optional[TStringValue] = None
    measurement_method_identifier: Optional[TStringValue] = None
    total_cycle_number_setting: Optional[TQuantityValueNumber] = None
    denaturing_temperature_setting: Optional[TQuantityValueDegreeCelsius] = None
    denaturing_time_setting: Optional[TQuantityValueSecondTime] = None
    annealing_temperature_setting: Optional[TQuantityValueDegreeCelsius] = None
    annealing_time_setting: Optional[TQuantityValueSecondTime] = None
    extension_temperature_setting: Optional[TQuantityValueDegreeCelsius] = None
    extension_time_setting: Optional[TQuantityValueSecondTime] = None
    PCR_detection_chemistry: Optional[TStringValue] = None
    reporter_dye_setting: Optional[TStringValue] = None
    quencher_dye_setting: Optional[TStringValue] = None
    passive_reference_dye_setting: Optional[TStringValue] = None


@dataclass
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass
class DataProcessingDocument:
    flourescence_intensity_threshold_setting: Optional[TQuantityValueUnitless] = None


@dataclass
class ProcessedDataDocumentItem:
    number_concentration: TQuantityValueNumberPerMicroliter
    positive_partition_count: TQuantityValueNumber
    data_processing_document: Optional[DataProcessingDocument] = None
    negative_partition_count: Optional[TQuantityValueNumber] = None


@dataclass
class ProcessedDataAggregateDocument:
    processed_data_document: list[ProcessedDataDocumentItem]


@dataclass
class MeasurementDocumentItem:
    measurement_identifier: TStringValue
    measurement_time: TDateTimeValue
    target_DNA_description: TStringValue
    sample_document: SampleDocument
    device_control_aggregate_document: DeviceControlAggregateDocument
    total_partition_count: TQuantityValueNumber
    processed_data_aggregate_document: ProcessedDataAggregateDocument


@dataclass
class MeasurementAggregateDocument:
    plate_well_count: TQuantityValueNumber
    measurement_document: list[MeasurementDocumentItem]
    analytical_method_identifier: Optional[TStringValue] = None
    experimental_data_identifier: Optional[TStringValue] = None
    experiment_type: Optional[ExperimentType] = None
    container_type: Optional[ContainerType] = None
    well_volume: Optional[TQuantityValueMicroliter] = None


@dataclass
class DataProcessingDocument1:
    reference_DNA_description: Optional[TStringValue] = None
    reference_DNA_copy_number_setting: Optional[TQuantityValueNumber] = None


@dataclass
class CalculatedDataDocumentItem:
    calculated_data_identifier: Optional[TStringValue] = None
    data_source_aggregate_document: Optional[DataSourceAggregateDocument] = None
    data_processing_document: Optional[DataProcessingDocument1] = None
    calculated_data_name: Optional[TStringValue] = None
    calculated_data_description: Optional[TStringValue] = None
    calculated_datum: Optional[TQuantityValue] = None


@dataclass
class TCalculatedDataAggregateDocument:
    calculated_data_document: Optional[list[CalculatedDataDocumentItem]] = None


@dataclass
class DPCRDocumentItem:
    measurement_aggregate_document: MeasurementAggregateDocument
    analyst: Optional[TStringValue] = None
    submitter: Optional[TStringValue] = None
    calculated_data_aggregate_document: Optional[
        TCalculatedDataAggregateDocument
    ] = None


@dataclass
class DPCRAggregateDocument:
    device_system_document: DeviceSystemDocument
    dPCR_document: list[DPCRDocumentItem]
    data_system_document: Optional[DataSystemDocument] = None
    calculated_data_aggregate_document: Optional[
        TCalculatedDataAggregateDocument
    ] = None


@dataclass
class Model:
    dPCR_aggregate_document: Optional[DPCRAggregateDocument] = None
    manifest: str = (
        "http://purl.allotrope.org/manifests/pcr/BENCHLING/2023/09/dpcr.manifest"
    )



from collections.abc import Iterator
from contextlib import contextmanager
from itertools import zip_longest
from pathlib import Path
import shutil
from typing import Optional, Union

PathType = Union[Path, str]


def _files_equal(path1: PathType, path2: PathType) -> bool:
    with open(str(path1)) as file1, open(str(path2)) as file2:
        for line1, line2 in zip_longest(file1, file2, fillvalue=""):
            if line1 != line2 and not line1.startswith("#   timestamp:"):
                return False
    return True


def _get_backup_path(path: PathType) -> Path:
    _path = Path(path)
    return Path(_path.parent, f".{_path.stem}.bak{_path.suffix}")


def is_file_changed(path: PathType) -> bool:
    backup_path = _get_backup_path(path)
    if backup_path.exists():
        return not _files_equal(path, backup_path)
    return True


def _backup_file(path: PathType) -> None:
    if Path(path).exists():
        shutil.copyfile(path, str(_get_backup_path(path)))


def restore_backup(path: PathType) -> None:
    backup_path = _get_backup_path(path)
    if backup_path.exists():
        backup_path.rename(path)


def _remove_backup(path: PathType) -> None:
    _get_backup_path(path).unlink(missing_ok=True)


@contextmanager
def backup(
    paths: Union[list[PathType], PathType], *, restore: Optional[bool] = False
) -> Iterator[None]:
    paths = paths if isinstance(paths, list) else [paths]
    for path in paths:
        _backup_file(path)
    try:
        yield
    except Exception:
        for path in paths:
            restore_backup(path)
        raise

    if restore:
        for path in paths:
            restore_backup(path)
    else:
        for path in paths:
            _remove_backup(path)



from __future__ import annotations

from enum import Enum
import tempfile
from typing import Any, Optional
import zipfile

import pandas as pd

from allotropy.allotrope.models.cell_counting_benchling_2023_11_cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDetectorDeviceControlAggregateDocument,
    CellCountingDetectorMeasurementDocumentItem,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlDocumentItemModel,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    Model,
    ProcessedDataAggregateDocument1,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_vi_cell_xr.constants import (
    DATE_HEADER,
    DEFAULT_ANALYST,
    MODEL_NUMBER,
    SOFTWARE_NAME,
    XrVersion,
)
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_reader import ViCellXRReader
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser
from allotropy.types import IOType


class SampleProperty(Enum):
    DILUTION_FACTOR = ("Dilution factor", TQuantityValueUnitless)
    TOTAL_CELLS_ML = ("Total cells/ml (x10^6)", TQuantityValueMillionCellsPerMilliliter)
    TOTAL_CELLS = ("Total cells", TQuantityValueCell)
    AVERAGE_DIAMETER = ("Avg. diam. (microns)", TQuantityValueMicrometer)
    VIABLE_CELLS = ("Viable cells", TQuantityValueCell)
    AVERAGE_CIRCULARITY = ("Avg. circ.", TQuantityValueUnitless)

    def __init__(self, column_name: str, data_type: Any) -> None:
        self.column_name: str = column_name
        self.data_type: Any = data_type


def get_property_from_sample(
    sample: pd.Series[Any], sample_property: SampleProperty
) -> Any:
    value = sample.get(sample_property.column_name)
    return sample_property.data_type(value=value) if value else None


def remove_style_xml_file(contents: IOType) -> IOType:
    # Removes styles.xml from an xlsx file IO stream. xlsx files produced by VI-Cell XR
    # instrument may have an invalid <fill> tag in their styles.xml file which causes a
    # bug when reading with pandas (via openpyxl library).

    # zipfile only accepts a filename, so write contents to a named temp file.
    tmp = tempfile.NamedTemporaryFile()
    file_contents = contents.read()
    if isinstance(file_contents, str):
        file_contents = file_contents.encode()
    tmp.write(file_contents)

    # Write zip contents to a new file, skipping styles.xml
    new = tempfile.NamedTemporaryFile()
    with zipfile.ZipFile(tmp.name) as zin:
        with zipfile.ZipFile(new.name, "w") as zout:
            for item in zin.infolist():
                if item.filename == "xl/styles.xml":
                    continue
                zout.writestr(item, zin.read(item.filename))

    return open(new.name, "rb")


class ViCellXRParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        contents = named_file_contents.contents
        filename = named_file_contents.original_file_name

        if filename.endswith("xlsx"):
            contents = remove_style_xml_file(contents)
        reader = ViCellXRReader(contents)

        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/11/cell-counting.manifest",
            cell_counting_aggregate_document=CellCountingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=MODEL_NUMBER,
                    equipment_serial_number=self._get_device_serial_number(
                        reader.file_info
                    ),
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name=SOFTWARE_NAME,
                    software_version=reader.file_version.value,
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                cell_counting_document=[
                    self._get_cell_counting_document_item(sample, reader.file_version)
                    for _, sample in reader.data.iterrows()
                ],
            ),
        )

    def _get_device_serial_number(self, file_info: pd.Series[Any]) -> Optional[str]:
        serial = str(file_info["serial"])
        try:
            serial_number = serial[serial.rindex(":") + 1 :].strip()
        except ValueError:
            return None

        return serial_number

    def _get_cell_counting_document_item(
        self, sample: pd.Series[Any], file_version: XrVersion
    ) -> CellCountingDocumentItem:
        required_fields_list = [
            "Viability (%)",
            "Viable cells/ml (x10^6)",
        ]
        # Required fields
        try:
            viability__cell_counter_ = TQuantityValuePercent(
                value=sample["Viability (%)"]
            )
            viable_cell_density__cell_counter_ = (
                TQuantityValueMillionCellsPerMilliliter(
                    value=sample["Viable cells/ml (x10^6)"]
                )
            )
        except KeyError as e:
            error = f"Expected to find lines with all of these headers: {required_fields_list}."
            raise AllotropeConversionError(error) from e

        return CellCountingDocumentItem(
            analyst=DEFAULT_ANALYST,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    CellCountingDetectorMeasurementDocumentItem(
                        measurement_identifier=random_uuid_str(),
                        measurement_time=self._get_date_time(
                            str(sample.get(DATE_HEADER[file_version]))
                        ),
                        sample_document=SampleDocument(
                            sample_identifier=sample["Sample ID"]
                        ),
                        device_control_aggregate_document=CellCountingDetectorDeviceControlAggregateDocument(
                            device_control_document=[
                                DeviceControlDocumentItemModel(
                                    device_type="brightfield imager (cell counter)",
                                    detection_type="brightfield",
                                )
                            ]
                        ),
                        processed_data_aggregate_document=ProcessedDataAggregateDocument1(
                            processed_data_document=[
                                ProcessedDataDocumentItem(
                                    data_processing_document=DataProcessingDocument(
                                        cell_type_processing_method=sample.get("Cell type"),  # type: ignore[arg-type]
                                        cell_density_dilution_factor=get_property_from_sample(
                                            sample,
                                            SampleProperty.DILUTION_FACTOR,
                                        ),
                                    ),
                                    viability__cell_counter_=viability__cell_counter_,
                                    viable_cell_density__cell_counter_=viable_cell_density__cell_counter_,
                                    total_cell_count=get_property_from_sample(
                                        sample, SampleProperty.TOTAL_CELLS
                                    ),
                                    total_cell_density__cell_counter_=get_property_from_sample(
                                        sample, SampleProperty.TOTAL_CELLS_ML
                                    ),
                                    average_total_cell_diameter=get_property_from_sample(
                                        sample, SampleProperty.AVERAGE_DIAMETER
                                    ),
                                    viable_cell_count=get_property_from_sample(
                                        sample, SampleProperty.VIABLE_CELLS
                                    ),
                                    average_total_cell_circularity=get_property_from_sample(
                                        sample, SampleProperty.AVERAGE_CIRCULARITY
                                    ),
                                ),
                            ]
                        ),
                    )
                ],
            ),
        )



from abc import ABC, abstractmethod
from typing import Any

from pandas import Timestamp

from allotropy.allotrope.models.shared.definitions.definitions import TDateTimeValue
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from allotropy.parsers.utils.values import assert_not_none


class VendorParser(ABC):
    timestamp_parser: TimestampParser

    """Base class for all vendor parsers."""

    def __init__(self, timestamp_parser: TimestampParser):
        self.timestamp_parser = assert_not_none(timestamp_parser, "timestamp_parser")

    @abstractmethod
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Any:
        raise NotImplementedError

    def _get_date_time(self, time: str) -> TDateTimeValue:
        assert_not_none(time, "time")

        return self.timestamp_parser.parse(time)

    # TODO(brian): Calling str() to pass to _get_date_time() potentially loses information.
    def _get_date_time_from_timestamp(self, timestamp: Timestamp) -> TDateTimeValue:
        # TODO(brian): fail if timestamp is not a Timestamp?

        assert_not_none(timestamp, "timestamp")

        time = str(timestamp)
        return self._get_date_time(time)



from abc import ABC, abstractmethod
from typing import Optional, Union

from allotropy.allotrope.models.plate_reader_benchling_2023_09_plate_reader import (
    ContainerType,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceSystemDocument,
    FluorescencePointDetectionDeviceControlAggregateDocument,
    FluorescencePointDetectionDeviceControlDocumentItem,
    FluorescencePointDetectionMeasurementDocumentItems,
    ImageDocumentItem,
    ImageFeatureAggregateDocument,
    ImageFeatureDocumentItem,
    LuminescencePointDetectionDeviceControlAggregateDocument,
    LuminescencePointDetectionDeviceControlDocumentItem,
    LuminescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    OpticalImagingAggregateDocument,
    OpticalImagingDeviceControlAggregateDocument,
    OpticalImagingDeviceControlDocumentItem,
    OpticalImagingMeasurementDocumentItems,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument,
    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMillimeter,
    TQuantityValueMilliSecond,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValuePercent,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueRelativeLightUnit,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.revvity_kaleido.kaleido_builder import create_data
from allotropy.parsers.revvity_kaleido.kaleido_structure import (
    Channel,
    Data,
    ExperimentType,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none
from allotropy.parsers.vendor_parser import VendorParser

MeasurementItem = Union[
    OpticalImagingMeasurementDocumentItems,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
    FluorescencePointDetectionMeasurementDocumentItems,
    LuminescencePointDetectionMeasurementDocumentItems,
]


class MeasurementParser(ABC):
    @abstractmethod
    def parse(self, data: Data, well_position: str) -> list[MeasurementItem]:
        pass

    @abstractmethod
    def get_experiment_type(self) -> ExperimentType:
        pass

    def get_sample_document(self, data: Data, well_position: str) -> SampleDocument:
        platemap_value = data.get_platemap_well_value(well_position)
        sample_identifier = (
            f"{data.results.barcode}_{well_position}"
            if platemap_value == "-"
            else platemap_value
        )
        return SampleDocument(
            sample_identifier=sample_identifier,
            location_identifier=str(well_position),
            well_plate_identifier=data.results.barcode,
            sample_role_type=data.get_sample_role_type(well_position),
        )


class FluorescenceMeasurementParser(MeasurementParser):
    def get_experiment_type(self) -> ExperimentType:
        return ExperimentType.FLUORESCENCE

    def get_device_control_document(
        self, data: Data
    ) -> FluorescencePointDetectionDeviceControlDocumentItem:
        return FluorescencePointDetectionDeviceControlDocumentItem(
            device_type="fluorescence detector",
            detection_type="fluorescence",
            number_of_averages=(
                None
                if data.measurements.number_of_averages is None
                else TQuantityValueNumber(value=data.measurements.number_of_averages)
            ),
            detector_distance_setting__plate_reader_=(
                None
                if data.measurements.detector_distance is None
                else TQuantityValueMillimeter(value=data.measurements.detector_distance)
            ),
            scan_position_setting__plate_reader_=data.measurements.scan_position,
            detector_wavelength_setting=(
                None
                if data.measurements.emission_wavelength is None
                else TQuantityValueNanometer(
                    value=data.measurements.emission_wavelength
                )
            ),
            excitation_wavelength_setting=(
                None
                if data.measurements.excitation_wavelength is None
                else TQuantityValueNanometer(
                    value=data.measurements.excitation_wavelength
                )
            ),
        )

    def parse(self, data: Data, well_position: str) -> list[MeasurementItem]:
        return [
            FluorescencePointDetectionMeasurementDocumentItems(
                measurement_identifier=random_uuid_str(),
                fluorescence=TQuantityValueRelativeFluorescenceUnit(
                    value=data.get_well_float_value(well_position)
                ),
                sample_document=self.get_sample_document(data, well_position),
                device_control_aggregate_document=FluorescencePointDetectionDeviceControlAggregateDocument(
                    device_control_document=[self.get_device_control_document(data)]
                ),
            )
        ]


class AbsorbanceMeasurementParser(MeasurementParser):
    def get_experiment_type(self) -> ExperimentType:
        return ExperimentType.ABSORBANCE

    def get_device_control_document(
        self, data: Data
    ) -> UltravioletAbsorbancePointDetectionDeviceControlDocumentItem:
        return UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
            device_type="absorbance detector",
            detection_type=data.background_info.experiment_type,
            number_of_averages=TQuantityValueNumber(
                value=assert_not_none(
                    data.measurements.number_of_averages,
                    msg="Unable to find number of averages",
                ),
            ),
            detector_distance_setting__plate_reader_=(
                None
                if data.measurements.detector_distance is None
                else TQuantityValueMillimeter(value=data.measurements.detector_distance)
            ),
            detector_wavelength_setting=(
                None
                if data.measurements.excitation_wavelength is None
                else TQuantityValueNanometer(
                    value=data.measurements.excitation_wavelength
                )
            ),
        )

    def parse(self, data: Data, well_position: str) -> list[MeasurementItem]:
        return [
            UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
                measurement_identifier=random_uuid_str(),
                absorbance=TQuantityValueMilliAbsorbanceUnit(
                    value=data.get_well_float_value(well_position)
                ),
                sample_document=self.get_sample_document(data, well_position),
                device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                    device_control_document=[self.get_device_control_document(data)]
                ),
            )
        ]


class LuminescenceMeasurementParser(MeasurementParser):
    def get_experiment_type(self) -> ExperimentType:
        return ExperimentType.LUMINESCENCE

    def get_device_control_document(
        self, data: Data
    ) -> LuminescencePointDetectionDeviceControlDocumentItem:
        return LuminescencePointDetectionDeviceControlDocumentItem(
            device_type="luminescence detector",
            detection_type=data.background_info.experiment_type,
            detector_distance_setting__plate_reader_=(
                None
                if data.measurements.detector_distance is None
                else TQuantityValueMillimeter(value=data.measurements.detector_distance)
            ),
        )

    def parse(self, data: Data, well_position: str) -> list[MeasurementItem]:
        return [
            LuminescencePointDetectionMeasurementDocumentItems(
                measurement_identifier=random_uuid_str(),
                luminescence=TQuantityValueRelativeLightUnit(
                    value=data.get_well_float_value(well_position)
                ),
                sample_document=self.get_sample_document(data, well_position),
                device_control_aggregate_document=LuminescencePointDetectionDeviceControlAggregateDocument(
                    device_control_document=[self.get_device_control_document(data)]
                ),
            )
        ]


class ImagingMeasurementParser(MeasurementParser):
    def get_experiment_type(self) -> ExperimentType:
        return ExperimentType.OPTICAL_IMAGING

    def get_device_control_document(
        self, data: Data, channel: Channel
    ) -> OpticalImagingDeviceControlDocumentItem:
        return OpticalImagingDeviceControlDocumentItem(
            device_type="imaging detector",
            detection_type="optical-imaging",
            detector_distance_setting__plate_reader_=(
                None
                if data.measurements.focus_height is None
                else TQuantityValueMillimeter(value=data.measurements.focus_height)
            ),
            excitation_wavelength_setting=(
                None
                if channel.excitation_wavelength is None
                else TQuantityValueNanometer(value=channel.excitation_wavelength)
            ),
            magnification_setting=TQuantityValueUnitless(value=4),
            exposure_duration_setting=(
                None
                if channel.exposure_duration is None
                else TQuantityValueMilliSecond(value=channel.exposure_duration)
            ),
            illumination_setting=(
                None
                if channel.illumination is None
                else TQuantityValuePercent(value=channel.illumination)
            ),
            transmitted_light_setting=channel.transmitted_light,
            fluorescent_tag_setting=channel.fluorescent_tag,
        )

    def parse(self, data: Data, well_position: str) -> list[MeasurementItem]:
        return [
            OpticalImagingMeasurementDocumentItems(
                measurement_identifier=random_uuid_str(),
                sample_document=self.get_sample_document(data, well_position),
                device_control_aggregate_document=OpticalImagingDeviceControlAggregateDocument(
                    device_control_document=[
                        self.get_device_control_document(data, channel)
                    ],
                ),
            )
            for channel in data.measurements.channels
        ]


class KaleidoParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        lines = read_to_lines(named_file_contents)
        reader = CsvReader(lines)
        data = create_data(reader)
        return self._get_model(named_file_contents.original_file_name, data)

    def _get_model(self, file_name: str, data: Data) -> Model:
        return Model(
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                plate_reader_document=self._get_plate_reader_document(data),
                device_system_document=self._get_device_system_document(data),
                data_system_document=self._get_data_system_document(
                    file_name, data.version
                ),
            ),
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
        )

    def _get_device_system_document(self, data: Data) -> DeviceSystemDocument:
        return DeviceSystemDocument(
            device_identifier="EnSight",
            model_number="EnSight",
            product_manufacturer="Revvity",
            equipment_serial_number=data.measurement_info.instrument_serial_number,
        )

    def _get_data_system_document(
        self, file_name: str, version: str
    ) -> DataSystemDocument:
        return DataSystemDocument(
            file_name=file_name,
            software_name="Kaleido",
            software_version=version,
            ASM_converter_name=ASM_CONVERTER_NAME,
            ASM_converter_version=ASM_CONVERTER_VERSION,
        )

    def _get_plate_reader_document(self, data: Data) -> list[PlateReaderDocumentItem]:
        measurement_parser = self._get_measurement_document_parser(
            data.background_info.experiment_type
        )
        return [
            PlateReaderDocumentItem(
                measurement_aggregate_document=self._get_measurement_aggregate_document(
                    data, measurement_parser, well_position
                )
            )
            for well_position in data.iter_wells()
        ]

    def _get_measurement_aggregate_document(
        self,
        data: Data,
        measurement_parser: MeasurementParser,
        well_position: str,
    ) -> MeasurementAggregateDocument:
        return MeasurementAggregateDocument(
            container_type=ContainerType.well_plate,
            measurement_time=self._get_date_time(
                data.measurement_info.measurement_time
            ),
            plate_well_count=TQuantityValueNumber(
                value=data.results.get_plate_well_count()
            ),
            experiment_type=data.background_info.experiment_type,
            analytical_method_identifier=data.measurement_info.protocol_signature,
            experimental_data_identifier=data.measurement_info.measurement_signature,
            measurement_document=measurement_parser.parse(data, well_position),
            image_aggregate_document=self.get_image_aggregate_document(
                data, well_position, measurement_parser
            ),
            processed_data_aggregate_document=self.get_processed_data_aggregate_document(
                data, well_position, measurement_parser
            ),
        )

    def get_image_aggregate_document(
        self,
        data: Data,
        well_position: str,
        measurement_parser: MeasurementParser,
    ) -> Optional[OpticalImagingAggregateDocument]:
        if measurement_parser.get_experiment_type() != ExperimentType.OPTICAL_IMAGING:
            return None

        return OpticalImagingAggregateDocument(
            image_document=[
                ImageDocumentItem(
                    experimental_data_identifier=data.get_well_str_value(well_position),
                )
            ]
        )

    def get_processed_data_aggregate_document(
        self,
        data: Data,
        well_position: str,
        measurement_parser: MeasurementParser,
    ) -> Optional[ProcessedDataAggregateDocument]:
        if (
            measurement_parser.get_experiment_type() != ExperimentType.OPTICAL_IMAGING
            or not data.analysis_results
        ):
            return None

        return ProcessedDataAggregateDocument(
            processed_data_document=[
                ProcessedDataDocumentItem(
                    image_feature_aggregate_document=ImageFeatureAggregateDocument(
                        image_feature_document=[
                            ImageFeatureDocumentItem(
                                image_feature_identifier=random_uuid_str(),
                                image_feature_name=analysis_result.analysis_parameter,
                                image_feature_result=TQuantityValueUnitless(
                                    value=analysis_result.get_image_feature_result(
                                        well_position
                                    )
                                ),
                                data_source_aggregate_document=DataSourceAggregateDocument(
                                    data_source_document=[
                                        DataSourceDocumentItem(
                                            data_source_identifier=data.get_well_str_value(
                                                well_position
                                            ),
                                            data_source_feature="image feature",  # wait for actual value
                                        )
                                    ]
                                ),
                            )
                            for analysis_result in data.analysis_results
                        ]
                    )
                )
            ]
        )

    def _get_measurement_document_parser(
        self, experiment_type: str
    ) -> MeasurementParser:
        experiment_type_lower = experiment_type.lower()
        if "fluorescence" in experiment_type_lower or "alpha" in experiment_type_lower:
            return FluorescenceMeasurementParser()
        elif "abs" in experiment_type_lower:
            return AbsorbanceMeasurementParser()
        elif "luminescence" in experiment_type_lower:
            return LuminescenceMeasurementParser()
        elif "img" in experiment_type_lower:
            return ImagingMeasurementParser()
        else:
            error = f"Unable to find valid experiment type in '{experiment_type}'"
            raise AllotropeConversionError(error)
